#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.operational contains functions to run the 
GNSS processing software RTKLIB. 

it can be imported directly with:
from geodezyx import operational

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""


import collections
########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
#### Import the logger
import logging
import os
import subprocess
import numpy as np

#### geodeZYX modules
from geodezyx import files_rw
from geodezyx import operational
from geodezyx import utils

log = logging.getLogger('geodezyx')

##########  END IMPORT  ##########


def read_conf_file(filein):
    outdic = collections.OrderedDict()
    for l in open(filein):
        if l[0] == "#":
            continue
        if l.strip() == "":
            continue

        f = l.split("=")
        key = f[0]
        if len(f) == 1:
            val = ""
        else:
            try:
                val = f[1].split("#")[0]
            except IndexError:
                val = f[1]
        outdic[key.strip()] = val.strip()
    return outdic


def rtklib_run_from_rinex(
    rnx_rover,
    rnx_base,
    generik_conf,
    working_dir,
    experience_prefix="",
    rover_auto_conf=False,
    base_auto_conf=True,
    XYZbase=[0, 0, 0],
    outtype="auto",
    calc_center="IGS0OPSFIN",
    exe_path = '/home/psakicki/SOFTWARE/RTKLIB_explorer/RTKLIB/app/consapp/rnx2rtkp/gcc/rnx2rtkp'
):
    """
    auto_conf :
        read the header of the rinex and write the conf. file according to it
        if the mode is disabled, the antenna/rec part of the
        conf. file will be the same as the generic one

        NB : RTKLIB "core" have it's own reading header option.
        Thus, my advice is to disable the auto mode for the rover and leave
        ``ant1-postype=rinexhead`` in the generic conf file
        and enable it for the base with a XYZbase vector initialized or
        the good XYZ in header on the rinex (the XYZbase vector is prioritary
        over the base RINEX header)

        (but in anycase, prepars good rinex's headers ;) )

    outtype :
        'auto' (as defined in the generic config file) or
        'dms' 'deg' 'xyz' 'enu'
        can manage the upper case XYZ or FLH
    """

    import shutil

    # paths & files
    working_dir = utils.create_dir(working_dir)
    out_dir = utils.create_dir(os.path.join(working_dir, "OUTPUT"))

    # paths & files
    temp_dir = os.path.join(working_dir,'TEMP')
    clean_temp_dir = False
    if clean_temp_dir:
        shutil.rmtree(temp_dir)
        temp_dir = os.path.join(working_dir,'TEMP')
    out_dir  = os.path.join(working_dir,'OUTPUT')

    # uncompressing rinex if compressed
    if operational.check_if_compressed_rinex(rnx_rover):
        rnx_rover = operational.crz2rnx(rnx_rover, temp_dir)
    if operational.check_if_compressed_rinex(rnx_base):
        rnx_base = operational.crz2rnx(rnx_base, temp_dir)

    # RINEX START & END
    rov_srt, rov_end, rov_itv = operational.rinex_start_end(rnx_rover, 1)
    bas_srt, bas_end, bas_itv = operational.rinex_start_end(rnx_base, 1)

    # RINEX NAMES
    rov_name = os.path.basename(rnx_rover)[0:4]
    bas_name = os.path.basename(rnx_base)[0:4]

    srt_str = rov_srt.strftime("%Y_%j")
    exp_full_name = "_".join((experience_prefix, rov_name, bas_name, srt_str))

    out_conf_fil = os.path.join(out_dir, exp_full_name + ".conf")
    out_result_fil = os.path.join(out_dir, exp_full_name + ".out")

    dicoconf = read_conf_file(generik_conf)

    if rover_auto_conf:
        antobj_rov, recobj_rov, siteobj_rov, locobj_rov = (
            files_rw.read_rinex_2_dataobjts(rnx_rover)
        )
        dicoconf["ant1-postype"] = "xyz"
        dicoconf["ant1-anttype"] = antobj_rov.Antenna_Type
        dicoconf["ant1-pos1"] = locobj_rov.X_coordinate_m
        dicoconf["ant1-pos2"] = locobj_rov.Y_coordinate_m
        dicoconf["ant1-pos3"] = locobj_rov.Z_coordinate_m
        dicoconf["ant1-antdelu"] = antobj_rov.Up_Ecc
        dicoconf["ant1-antdeln"] = antobj_rov.North_Ecc
        dicoconf["ant1-antdele"] = antobj_rov.East_Ecc

    if not outtype.lower() == "auto":
        dicoconf["out-solformat"] = outtype.lower()
        log.info("out-solformat", dicoconf["out-solformat"])

    if base_auto_conf:
        antobj_bas, recobj_bas, siteobj_bas, locobj_bas = (
            files_rw.read_rinex_2_dataobjts(rnx_base)
        )
        dicoconf["ant2-postype"] = "xyz"
        dicoconf["ant2-anttype"] = antobj_bas.Antenna_Type
        if XYZbase[0] != 0:
            dicoconf["ant2-pos1"] = XYZbase[0]
            dicoconf["ant2-pos2"] = XYZbase[1]
            dicoconf["ant2-pos3"] = XYZbase[2]
        else:
            dicoconf["ant2-pos1"] = locobj_bas.X_coordinate_m
            dicoconf["ant2-pos2"] = locobj_bas.Y_coordinate_m
            dicoconf["ant2-pos3"] = locobj_bas.Z_coordinate_m
        dicoconf["ant2-antdelu"] = antobj_bas.Up_Ecc
        dicoconf["ant2-antdeln"] = antobj_bas.North_Ecc
        dicoconf["ant2-antdele"] = antobj_bas.East_Ecc

    if not (bas_srt <= rov_srt <= rov_end <= bas_end):
        log.warning("not bas_srt <= rov_srt <= rov_end <= bas_end !!!")

    outconffilobj = open(out_conf_fil, "w+")
    for k, v in dicoconf.items():
        lin = k.ljust(20) + "=" + str(v) + "\n"
        outconffilobj.write(lin)
    outconffilobj.close()

    ###### ORBITS
    ### SP3
    orblis = operational.download_gnss_products(
        archive_dir=temp_dir,
        startdate=bas_srt,
        enddate=bas_end,
        archtype="/",
        AC_names=(calc_center , )
    )

    unzip = lambda o: files_rw.unzip_gz_z(o) if o.endswith(".gz") or o.endswith(".Z") else o

    sp3_z = orblis[0]
    sp3 = unzip(sp3_z)

    ### BRDC
    statdic = dict()
    statdic["nav"] = ["BRDC"]
    nav_srt = dt.datetime(bas_srt.year, bas_srt.month, bas_srt.day)
    brdclis = operational.download_gnss_rinex(
        statdic, temp_dir, nav_srt, bas_end, archtype="/"
    )

    if len(brdclis) > 0 and brdclis[0][1]:
        nav_z = brdclis[0][0]
        nav = unzip(nav_z)

    else:
        log.error("No BRDC nav file found in the archive")
        raise FileNotFoundError("No BRDC nav file found in the archive")

    # Command
    com_config = "-k " + out_conf_fil
    com_interval = "-ti " + str(np.round(rov_itv,6))
    com_mode = ""
    # com_mode="-p 4"
    com_resultfile = "-o " + out_result_fil
    # com_combinsol="-c"

    # exe_path = "rnx2rtkp"
    #    exe_path = "/home/pierre/install_softs/RTKLIB/rnx2rtkp"
    # exe_path = "/home/psakicki/SOFTWARE/RTKLIB/RTKLIB/app/rnx2rtkp/gcc/rnx2rtkp"
    # exe_path = ""

    bigcomand = " ".join(
        (
            exe_path,
            com_config,
            com_interval,
            com_mode,
            com_resultfile,
            rnx_rover,
            rnx_base,
            nav,
            sp3,
        )
    )

    log.info(bigcomand)

    subprocess.call([bigcomand], executable="/bin/bash", shell=True)
    log.info("RTKLIB RUN FINISHED")

    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 17 17:41:32 2022

@author: psakic

This module regroups the functions for the resolution estimation of the 
REVOSIMA's OBSCOM-embeded pressure sensor
"""

import itertools
#### Import the logger
import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy

import geodezyx.conv as conv
from geodezyx import utils

log = logging.getLogger('geodezyx')


#  ______               _                 _    __                  _   _
# |  ____|             | |               | |  / _|                | | (_)
# | |__ _ __ ___  _ __ | |_ ___ _ __   __| | | |_ _   _ _ __   ___| |_ _  ___  _ __  ___
# |  __| '__/ _ \| '_ \| __/ _ \ '_ \ / _` | |  _| | | | '_ \ / __| __| |/ _ \| '_ \/ __|
# | |  | | | (_) | | | | ||  __/ | | | (_| | | | | |_| | | | | (__| |_| | (_) | | | \__ \
# |_|  |_|  \___/|_| |_|\__\___|_| |_|\__,_| |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|___/


def obscom_import_pres_df(
        csv_pres_inp,
        dic_coeffs,
        freq_clk=4.096e6,
        count_sensor_temp=168e3,
        count_sensor_pres=30e3,
        integ_timespan_temp=10,
        integ_timespan_pres=100,
        epoc_idx=True,
        out_pres="pa",
):
    """
    Imports pressure data from a CSV file and computes the frequency, temperature, pressure, and depth.

    Parameters
    ----------
    csv_pres_inp : str
        The path to the CSV file containing the pressure data.
    dic_coeffs : dict
        The dictionary of coefficients used for the computations.
    freq_clk : float, optional
        The frequency of the clock in Hertz. Default is 4.096e6.
    count_sensor_temp : float, optional
        The count of the temperature sensor. Default is 168e3.
    count_sensor_pres : float, optional
        The count of the pressure sensor. Default is 30e3.
    integ_timespan_temp : int, optional
        The integration timespan for the temperature sensor. Default is 10.
    integ_timespan_pres : int, optional
        The integration timespan for the pressure sensor. Default is 100.
    epoc_idx : bool, optional
        If True, sets the epoch as the index of the DataFrame. Default is True.
    out_pres : str, optional
        The unit of the output pressure. Default is "pa".

    Returns
    -------
    DataFrame
        A DataFrame with the epoch, frequency, temperature, pressure, and depth.

    Notes
    -----
    This function first reads the CSV file and creates a DataFrame.
    It then computes the frequency from the counter for both the temperature and pressure sensors.
    The temperature is computed from the frequency using the coefficients from the dictionary.
    The pressure is computed from the frequency of both the pressure and temperature sensors.
    The depth is also computed from the frequency of both the pressure and temperature sensors.
    If epoc_idx is True, the epoch is set as the index of the DataFrame.
    """

    dic_coeffs_temp = _get_coeffs_temp(dic_coeffs)

    pres_df_raw = pd.read_csv(csv_pres_inp)

    pres_df = pd.DataFrame()

    pres_df["epoc"] = conv.posix2dt(pres_df_raw["Timestamp(unixtime)"])

    ### rename columns
    pres_df["cnt_temp"] = pres_df_raw["Temperature"]
    pres_df["cnt_pres"] = pres_df_raw["Pressure"]

    ### remove empty values
    pres_df.replace(0, np.nan, inplace=True)
    pres_df.dropna(inplace=True)

    ### compute frequency from counter
    pres_df["frq_temp"] = counter2freq(
        pres_df["cnt_temp"],
        count_sensor=count_sensor_temp,
        freq_clk=freq_clk,
        integ_timespan=integ_timespan_temp,
    )

    pres_df["frq_pres"] = counter2freq(
        pres_df["cnt_pres"],
        count_sensor=count_sensor_pres,
        freq_clk=freq_clk,
        integ_timespan=integ_timespan_pres,
    )
    ### compute temperature from frequency
    pres_df["temp"] = freq2temp(pres_df["frq_temp"], **dic_coeffs_temp)

    ### compute pressure from frequency, using temperature frequency
    pres_df["pres"] = freq2pres(
        pres_df["frq_pres"],
        pres_df["frq_temp"],
        out=out_pres,
        inp_temp="freq",
        **dic_coeffs
    )

    ### compute pressure from frequency, using temperature itself
    ### (for debug only!)
    debug = False
    if debug:
        pres_df["pres2"] = freq2pres(pres_df["frq_pres"],
                                     pres_df["temp"],
                                     out="pa",
                                     inp_temp="temp",
                                     **dic_coeffs)

    ### compute depth
    pres_df["dept"] = freq2pres(
        pres_df["frq_pres"], pres_df["temp"], out="meter", **dic_coeffs
    )

    if epoc_idx:
        pres_df.set_index("epoc", inplace=True)

    return pres_df


def obscom_plot_pres_df(pres_df_in, plot_count=False):
    """
    Plots the pressure and temperature data from a DataFrame.

    Parameters
    ----------
    pres_df_in : DataFrame
        The DataFrame containing the pressure and temperature data. The index of the DataFrame should be the time.
    plot_count : bool, optional
        If True, plots the count of the temperature and pressure sensors. If False, plots the pressure in pa and temperature in °C. Default is False.

    Returns
    -------
    Figure
        The Figure object with the plot of the pressure and temperature data.

    Notes
    -----
    This function creates a plot with two y-axes. The left y-axis corresponds to the pressure data and the right y-axis corresponds to the temperature data.
    The x-axis corresponds to the time.
    The function uses different colors for the pressure and temperature data for better distinction.
    """

    fig, ax1 = plt.subplots()

    if not plot_count:
        col1 = 'pres'
        col2 = 'temp'
        lgd1 = 'pressure (pa)'
        lgd2 = 'temperature (°C)'

    else:
        col1 = 'cnt_temp'
        col2 = 'cnt_pres'
        lgd1 = 'count (#)'
        lgd2 = 'count (#)'

    ax1.set_xlabel('time')
    ax1.set_ylabel(lgd1, color='C1')
    ax1.plot(pres_df_in.index, pres_df_in[col1], color='C1')
    ax1.tick_params(axis='y', labelcolor='C1')

    # Adding Twin Axes
    ax2 = ax1.twinx()
    ax2.plot(pres_df_in.index, pres_df_in[col2], color='C2')
    ax2.set_ylabel(lgd2, color='C2')
    ax2.tick_params(axis='y', labelcolor='C2')

    return fig


#                  __  __ _      _            _                                      _             _
#                 / _|/ _(_)    (_)          | |         ___                        | |           | |
#   ___ ___   ___| |_| |_ _  ___ _  ___ _ __ | |_ ___   ( _ )     ___ ___  _ __  ___| |_ ___ _ __ | |_ ___
#  / __/ _ \ / _ \  _|  _| |/ __| |/ _ \ '_ \| __/ __|  / _ \/\  / __/ _ \| '_ \/ __| __/ _ \ '_ \| __/ __|
# | (_| (_) |  __/ | | | | | (__| |  __/ | | | |_\__ \ | (_>  < | (_| (_) | | | \__ \ ||  __/ | | | |_\__ \
#  \___\___/ \___|_| |_| |_|\___|_|\___|_| |_|\__|___/  \___/\/  \___\___/|_| |_|___/\__\___|_| |_|\__|___/
#

## Wikipedia values
#  https://en.wikipedia.org/wiki/Pound_per_square_inch

PSI_PER_METER = 1.4503773773022  ## 1 METER = 1 DBAR
PSI_TO_BAR = 0.06894757293168
PSI_TO_MBAR = PSI_TO_BAR * 1000
PSI_TO_PA = 6894.757293168


def get_coeffs(sens_type="all", sens_id=158073):
    if sens_id == 0:
        ### from exemple sheets
        dic_coeffs = {
            "U0": 5.766353,
            "Y1": -4025.183,
            "Y2": -11970.76,
            "Y3": 0.0,
            "T1": 29.89307,
            "T2": 0.337614,
            "T3": 56.99511,
            "T4": 157.6942,
            "T5": 0.0,
            "C1": -22682.65,
            "C2": -1143.743,
            "C3": 70903.62,
            "D1": 0.040903,
            "D2": 0.0,
        }

    elif sens_id == 158073 or sens_id == "T3":
        ### for the 1st OBSCOM
        dic_coeffs = {
            "U0": 5.798999,
            "Y1": -3874.954,
            "Y2": -10166.55,
            "Y3": 0,
            "C1": -25657.25,
            "C2": -645.8024,
            "C3": 73515.96,
            "D1": 0.039737,
            "D2": 0,
            "T1": 30.00177,
            "T2": 0.723913,
            "T3": 53.84611,
            "T4": 147.1236,
            "T5": 0,
        }

    elif sens_id == 158073999 or sens_id == "T3gross":  ### 999 suffix = gross
        ### for the 1st OBSCOM
        dic_coeffs = {
            "U0": 5.799,
            "Y1": -3874.95,
            "Y2": -10166.5,
            "Y3": 0,
            "C1": -25657.2,
            "C2": -645.802,
            "C3": 73516,
            "D1": 0.0397368,
            "D2": 0,
            "T1": 30.0018,
            "T2": 0.723913,
            "T3": 53.8461,
            "T4": 147.124,
            "T5": 0,
        }

    elif sens_id == 158076 or sens_id == "T2":
        ### for the 2nd OBSCOM (T2)
        dic_coeffs = {
            "U0": 5.765270,
            "Y1": -3994.585,
            "Y2": -10705.43,
            "Y3": 0,
            "C1": -25561.86,
            "C2": -230.0508,
            "C3": 79516.88,
            "D1": 0.040172,
            "D2": 0,
            "T1": 30.13945,
            "T2": 0.977439,
            "T3": 56.64988,
            "T4": 158.4542,
            "T5": 0,
        }

    elif sens_id == 158076999 or sens_id == "T2gross":  ### 999 suffix = gross
        ### for the 2nd OBSCOM (T2)
        dic_coeffs = {
            "U0": 5.76527,
            "Y1": -3994.58,
            "Y2": -10705.4,
            "Y3": 0,
            "C1": -25561.9,
            "C2": -230.051,
            "C3": 79516.9,
            "D1": 0.0401721,
            "D2": 0,
            "T1": 30.1394,
            "T2": 0.977439,
            "T3": 56.6499,
            "T4": 158.454,
            "T5": 0,
        }

    elif sens_id == 158081 or sens_id == "T4":
        ### for the 2nd OBSCOM (T2)
        dic_coeffs = {
            "U0": 5.797503,
            "Y1": -4022.350,
            "Y2": -12679.85,
            "Y3": 0,
            "C1": -22175.21,
            "C2": -1252.800,
            "C3": 83574.76,
            "D1": 0.037913,
            "D2": 0,
            "T1": 30.23989,
            "T2": 0.124474,
            "T3": 64.95771,
            "T4": 229.3125,
            "T5": 0,
        }

    if sens_type == "all":
        return dic_coeffs
    elif sens_type == "f0":
        return _get_coeffs_f0(dic_coeffs)
    elif sens_type == "temp":
        return _get_coeffs_temp(dic_coeffs)
    else:
        log.error("check sens_type ('pres' or 'temp')")


def _get_coeffs_temp(dic_coeffs_in):
    """
    return only the useful coefficients to compute temp
    """

    return {c: dic_coeffs_in[c] for c in ["U0", "Y1", "Y2", "Y3"]}


def _get_coeffs_f0(dic_coeffs_in):
    """
    return only the useful coefficients to compute f0
    """

    return {c: dic_coeffs_in[c] for c in ["U0", "T1", "T2", "T3", "T4", "T5"]}


def unitary_tests(sens_id=158073):
    """
    return the correct values for pressure/temperature and
    intermediates parameters @ ftemp=172600.0 Hz and fpres=36300.0 Hz
    with the cofficients of the sensor #158073
    (1st OBSCOM)

    Returns
    -------
    ftemp (Hz), fpres (Hz), temp (°C), f0 (Hz), p (psi)
    """

    if sens_id == 158073:
        ftemp = 172600.0  # Hz
        fpres = 36300.0  # Hz
        temp = 20.090562800024895  # °C
        f0 = 33333.93215844317  # Hz # @ 20.090562800024895°C
        p = 4803.3285794411595  # psi # @ 20.090562800024895 °C
    return ftemp, fpres, temp, f0, p


#  _                                      _
# | |                                    | |
# | |_ ___ _ __ ___  _ __   ___ _ __ __ _| |_ _   _ _ __ ___
# | __/ _ \ '_ ` _ \| '_ \ / _ \ '__/ _` | __| | | | '__/ _ \
# | ||  __/ | | | | | |_) |  __/ | | (_| | |_| |_| | | |  __/
#  \__\___|_| |_| |_| .__/ \___|_|  \__,_|\__|\__,_|_|  \___|
#                   | |
#                   |_|


def freq2U(ft_in, U0):
    """
    Convert temperature sensor frequency to coefficient U

    Parameters
    ----------
    ft_in : float
        frequency of the temperature sensor in Hertz.
    U0 : float
        input coefficient U0.

    Returns
    -------
    float
        coefficient U.

    """
    return 1e6 / ft_in - U0


def freq2temp(ft_in, U0, Y1, Y2, Y3):
    """
    Convert temperature sensor frequency to temperature

    Parameters
    ----------
    ft_in :
        frequency of the temperature sensor in Hertz.
    U0, Y1-3 : float
        Calibration coefficients provided by PAROS.

    Returns
    -------
    float
        temperature in Celsius.

    """

    if utils.is_iterable(ft_in):
        return np.array([freq2temp(e, U0, Y1, Y2, Y3) for e in ft_in])
    else:
        if ft_in == 0.0:
            ft_in = np.nan

        # X = (1/ft_in) * 10**6 # (1/30000) * 10**6
        # U = X - U0

        U = freq2U(ft_in, U0)

        Temp = Y1 * U + Y2 * U ** 2 + Y3 * U ** 3

        return Temp


def temp2freq(t_in, U0, Y1, Y2, Y3, out="freq", force_root=False):
    """
    Convert temperature to temperature sensor frequency

    Parameters
    ----------
    t_in :
        temperature in Celsius.
    U0, Y1-3 : float, optional
        Calibration coefficients provided by PAROS.
    out : float, optional
        freqency ('freq') or period ('tau') or coefficient U ('U')
        of the sensor. The default is 'freq'.
    force_root : bool or int
        force the finded root for U
        if integer (1 or 2), use the first or second root
        default is False

    Returns
    -------
    float
        temperature sensor frequency or period.

    Note
    ----
    This function is unstable because of the root finding

    You can force it with `force_root`

    """

    #### Handle T as an interable (array, list....)
    if utils.is_iterable(t_in):
        return np.array([temp2freq(t, U0, Y1, Y2, Y3, out, force_root) for t in t_in])
    #### Handle T as a scalar
    else:
        Uraw = np.roots([Y3, Y2, Y1, -t_in]) + U0
        U = Uraw
        F = 10 ** 6 * U ** -1  ### converted to frequency

        # Fbool = np.logical_and((F < 176000) , (F > 168000))

        Fok = np.min(F)
        Uok = U[Fok == F][0]

        if False and not ((10 ** 6 / U < 176000).all() and (10 ** 6 / U > 168000).all()):
            log.warn("temp U/freq out of range: %f/%f", U, 10 ** 6 / U)

        if out == "freq":
            #            return [ 10**6/u  for u in U  if ( (10**6/u< 176000) and (10**6/u > 168000 ))][0]
            # return [ 10**6/u  for u in U ][ilist]
            # outvals = 10**6 / U
            outvals = Fok
        elif out == "tau":
            # return [ 10**6/u  for u in U  if ( (10**6/u< 176000) and (10**6/u > 168000 ))][ilist]**-1
            # return [ 10**6/u  for u in U ][ilist]**-1
            # outvals = U / 10**6
            outvals = Fok ** -1
        elif out == "U":
            #            return [ u-U0  for u in U  if ( (u> 10**6/176000) and (u < 10**6/168000 ))][0]
            # return [ u-U0  for u in U ][ilist]
            # outvals = Uok - U0
            outvals = freq2U(Fok, U0)
        else:
            log.error("bad out value")
            raise Exception("bad output format")

    return outvals


def _temp2pres_f0(t_or_ft_in, U0, Y1, Y2, Y3, T1, T2, T3, T4, T5, inp_temp="temp"):
    """
    compute the f0 value, i.e. the temperature component for
    the pressure computation

    f0 is homogeneous to a frequency

    Internal function for ``freq2pres``

    Parameters
    ----------
    t_or_ft_in : float
        temperature in Celsius or temperature sensor frequency in Hertz.
        depends on ``inp_temp``
    U0, T1-5 : float, optional
        Calibration coefficients provided by PAROS.
    inp_temp : str
        'freq' temperature sensor frequency in Hertz
        'temp' temperature in Celsius
        default is 'temp'

    Returns
    -------
    float
        f0, the temperature component for the pressure computation.

    Note
    ----
    Using temperature as input is unstable because of the root finding for U

    frequency as input is recommended
    """
    if inp_temp == "temp":
        F = temp2freq(t_or_ft_in, U0, Y1, Y2, Y3, out="freq")
    else:
        F = t_or_ft_in

    # U= (1/F)*10**6 - U0
    U = freq2U(F, U0)

    T0 = T1 + T2 * U + T3 * U ** 2 + T4 * U ** 3 + T5 * U ** 4

    F0 = 10 ** 6 / T0

    return F0


#  _ __  _ __ ___  ___ ___ _   _ _ __ ___
# | '_ \| '__/ _ \/ __/ __| | | | '__/ _ \
# | |_) | | |  __/\__ \__ \ |_| | | |  __/
# | .__/|_|  \___||___/___/\__,_|_|  \___|
# | |
# |_|


def freq2pres(
        fp_in,
        t_or_ft_in,
        C1,
        C2,
        C3,
        D1,
        D2,
        U0,
        Y1,
        Y2,
        Y3,
        T1,
        T2,
        T3,
        T4,
        T5,
        out="psi",
        inp_temp="temp",
):
    """
    Convert pressure sensor frequency to pressure or equivalent distance

    Parameters
    ----------
    fp_in :
        frequency of the pressure sensor in Hertz.
    t_or_ft_in : float
        temperature in Celsius or temperature sensor frequency in Hertz.
        depends on ``inp_temp``
    C1 C2 C3 D1 D2 : float
        Calibration coefficients provided by PAROS.
    U0,Y1,Y2,Y3,T1,T2,T3,T4,T5 : float
        Calibration coefficients provided by PAROS for the temperature sensor.
    out : str, optional
        output as pressure ('psi', 'bar', 'mbar', 'pa') or distance ('meter').
        The default is "psi".
    inp_temp : str
        'freq' temperature sensor frequency in Hertz,
        'temp' temperature in Celsius,
        default is 'temp'

    Returns
    -------
    float
        pressure or equivalent distance.

    Note
    ----
    Using temperature as input is unstable because of the root finding of U
    frequency as input is recommended

    """

    if inp_temp == "temp":
        U = temp2freq(t_or_ft_in, U0, Y1, Y2, Y3, out="U")
    elif inp_temp == "freq":
        U = freq2U(t_or_ft_in, U0)

    C = C1 + C2 * U + C3 * U ** 2
    D = D1 + D2 * U

    T = (1 / fp_in) * 10 ** 6

    ### takes inp_temp = 'freq' or 'temp'
    F0 = _temp2pres_f0(
        t_or_ft_in, U0, Y1, Y2, Y3, T1, T2, T3, T4, T5, inp_temp=inp_temp
    )
    T0 = (1 / F0) * 10 ** 6

    P = C * (1 - (T0 ** 2) / (T ** 2)) * (1 - D * (1 - (T0 ** 2) / (T ** 2)))

    if out == "psi":
        return P
    elif out == "bar":
        return P * PSI_TO_BAR
    elif out == "mbar":
        return P * PSI_TO_MBAR
    elif out == "pa":
        return P * PSI_TO_PA
    elif out == "meter":
        return P / PSI_PER_METER
    else:
        log.error("bad out value")
        raise Exception("bad output format")


def pres2freq(
        p_in,
        t_in,
        C1,
        C2,
        C3,
        D1,
        D2,
        U0,
        Y1,
        Y2,
        Y3,
        T1,
        T2,
        T3,
        T4,
        T5,
        inp="psi",
        return_optimize_object=False,
):
    """
    Convert pressure to  pressure sensor frequency

    There is no analytic solution, thus an root find is required

    Parameters
    ----------
    p_in :
        pressure in psi.
    t_in :
        temperature in Celsius.
    inp : optional
        input for p_in as pressure ('psi', 'bar', 'mbar', 'pa')
        or distance ('meter').
        The default is 'psi'.
    return_optimize_object : bool, optional
        if True, return the object of scipy's optimize.root function.
        if False, return the root value
        The default is False.

    Returns
    -------
    float
        pressure sensor frequency (Hertz).

    Note
    ----

    """

    #### Handle p_in as an interable (array, list....)
    if utils.is_iterable(t_in):
        log.err("t_in has to be a scalar")
        raise Exception

    if utils.is_iterable(p_in):
        return np.array(
            [
                pres2freq(
                    p_in_i,
                    t_in,
                    C1,
                    C2,
                    C3,
                    D1,
                    D2,
                    U0,
                    Y1,
                    Y2,
                    Y3,
                    T1,
                    T2,
                    T3,
                    T4,
                    T5,
                    inp,
                    return_optimize_object,
                )
                for p_in_i in p_in
            ]
        )

    #### Handle p_in as a scalar
    else:
        if inp == "psi":
            puse = p_in
        elif inp == "bar":
            puse = p_in / PSI_TO_BAR
        elif inp == "mbar":
            puse = p_in / PSI_TO_MBAR
        elif inp == "pa":
            puse = p_in / PSI_TO_PA
        elif inp == "meter":
            puse = p_in * PSI_PER_METER

        def wrap_zero(fp):
            return (
                    freq2pres(
                        fp, t_in, C1, C2, C3, D1, D2, U0, Y1, Y2, Y3, T1, T2, T3, T4, T5
                    )
                    - puse
            )

        Opti = scipy.optimize.root(wrap_zero, 30000)

        if return_optimize_object:
            return Opti
        else:
            return Opti.x[0]


#                        _
#                       | |
#   ___ ___  _   _ _ __ | |_ ___ _ __
#  / __/ _ \| | | | '_ \| __/ _ \ '__|
# | (_| (_) | |_| | | | | ||  __/ |
#  \___\___/ \__,_|_| |_|\__\___|_|


def freq2counter(
        freq_or_tau_sensor,
        count_sensor,
        freq_or_tau_clk,
        integ_timespan=1,
        inp="freq",
        round_fct=np.floor,
):
    """
    Convert sensor frequency to count number

    Parameters
    ----------
    freq_or_tau_sensor :
        freqency ('freq') or period ('tau') of the sensor.
    count_sensor : int or float
        number of clock periods counted for one sensor count
    freq_or_tau_clk :
        primary frequency/period of the clock that count period..
    integ_timespan : int or float, optional
        integration timespan in seconds. The default is 1.
    inp : str, optional
        input 'freq' for freqency or 'tau' for period (=1/frequency).
        The default is 'freq'.
    round_fct : function, optional
        the function that round the count values. The default is np.floor.

    Returns
    -------
    n_count :
        number of counts counted by the sensor..

    Note
    ----
    usually, ``count_sensor=30e3``  for pressure, ``count_sensor=168e3``
    for temperature and ``freq_clk=4.096e6``

    """

    if inp == "freq":
        tau_sensor = 1 / freq_or_tau_sensor
        tau_clk = 1 / freq_or_tau_clk
    elif inp == "tau":
        tau_sensor = freq_or_tau_sensor
        tau_clk = freq_or_tau_clk
    else:
        log.error("bad inp value")
        raise Exception("bad inp format")

    n_count = (integ_timespan * tau_sensor * count_sensor) / tau_clk

    if round_fct:
        return round_fct(n_count)
    else:
        return n_count


def counter2freq(
        n_counted_by_clk, count_sensor, freq_clk, integ_timespan=1, out="freq"
):
    """
    Convert count number to sensor frequency

    Parameters
    ----------
    n_counted_by_clk : int or float
        number of counts counted by the sensor.
    count_sensor : int or float
        number of clock periods counted for one sensor count
    freq_clk : int or float
        primary frequency of the clock that count period.
    integ_timespan : int or float, optional
        integration timespan in seconds. The default is 1.
    out : str, optional
        output 'freq' for freqency or 'tau' for period (=1/frequency).
        The default is 'freq'.

    Returns
    -------
    float
        freqency ('freq') or period ('tau') of the sensor

    Note
    ----
    usually, ``count_sensor=30e3``  for pressure, ``count_sensor=168e3``
    for temperature and ``freq_clk=4.096e6``

    """

    tau_sensor = n_counted_by_clk / (freq_clk * integ_timespan * count_sensor)

    if out == "freq":
        return 1 / tau_sensor
    elif out == "tau":
        return tau_sensor
    else:
        log.error("bad out value")
        raise Exception("bad output format")


####### HIGHER LEVEL FCTS #####################################################


def pres_resolution(
        val_presin,
        val_temp_in,
        val_clkin,
        count_presin,
        count_temp_in,
        input_val="freq",
        output_val="meter",
        relative_delta=True,
):
    if input_val == "freq":
        if np.any(val_temp_in < 1) or np.any(val_presin < 1) or np.any(val_clkin < 1):
            log.warn("some vals are < 1 while input is freq. Please double check")
        tau_temp = 1 / val_temp_in
        tau_pres = 1 / val_presin
        tau_clk = 1 / val_clkin
    elif input_val == "tau":
        if np.any(val_temp_in > 1) or np.any(val_presin > 1) or np.any(val_clkin > 1):
            log.warn("some vals are > 1 while input is tau. Please double check")
        tau_temp = val_temp_in
        tau_pres = val_presin
        tau_clk = val_clkin

    # internal values are tau, because at least it simplifies the computation
    # of pres_bias  and avoids the 1/(a+b) = 1/a + 1/b mistake

    freq_clk = 1 / tau_clk

    temp_central = freq2temp(1 / tau_temp)
    pres_central = freq2pres(1 / tau_pres, temp_central)

    tau_pres_bias = 2 / (freq_clk * count_presin)
    tau_temp_bias = 2 / (freq_clk * count_temp_in)

    PresResArr = np.zeros((3, 3))

    if relative_delta:
        divider = pres_central
    else:
        divider = 1.0

    for i, j in [(1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0), (-1, 1), (1, 1), (0, 1)]:
        pres_biased = freq2pres(
            1 / (tau_pres + i * tau_pres_bias),
            freq2temp(1 / (tau_temp + j * tau_temp_bias)),
        )

        pres_res = np.abs(pres_biased - pres_central) / divider

        PresResArr[i + 1, j + 1] = pres_res

    if output_val == "psi":
        pass
    elif output_val == "meter":
        pres_central = pres_central / PSI_PER_METER
        if not relative_delta:
            PresResArr = (
                    PresResArr / PSI_PER_METER
            )  ### correction only for absolute values !!!
    else:
        raise Exception("bad output format in pres_resolution")

    return pres_central, temp_central, PresResArr.max(), PresResArr


def resolution_grid_compute(
        Tau_pres,
        Tau_temp,
        fe,
        count_pres,
        count_temp,
        output_val="meter",
        relative_delta=False,
):
    PresVals = np.zeros((len(Tau_pres), len(Tau_temp)))
    TempVals = np.zeros((len(Tau_pres), len(Tau_temp)))
    ResVals = np.zeros((len(Tau_pres), len(Tau_temp)))
    FullResVals = np.zeros((len(Tau_pres), len(Tau_temp), 3, 3))

    for itaupres, itautemp in itertools.product(
            range(len(Tau_pres)), range(len(Tau_temp))
    ):
        taupres, tautemp = Tau_pres[itaupres], Tau_temp[itautemp]

        OutRes = pres_resolution(
            taupres,
            tautemp,
            1 / fe,
            count_pres,
            count_temp,
            input_val="tau",
            output_val=output_val,
            relative_delta=relative_delta,
        )

        PresVals[itaupres, itautemp] = OutRes[0]
        TempVals[itaupres, itautemp] = OutRes[1]
        ResVals[itaupres, itautemp] = OutRes[2]
        FullResVals[itaupres, itautemp, :, :] = OutRes[3]

    return PresVals, TempVals, ResVals, FullResVals


############ PLOT OF THE RESOLUTION VALUES
def resolution_plot_as_gradient_grid(
        PresVals,
        TempVals,
        ResVals,
        temp_ref,
        freq_input=False,
        coef_res=1000,
        relative_delta=False,
):
    #### coef_res = 10**6 if PPM
    #### coef_res = 10**3 if error in meters

    #################### INITIALISATION ##########################

    secx_pres2freq = lambda x: pres2freq(x, temp_ref, "meter")
    secx_freq2pres = lambda x: freq2pres(x, temp_ref, "meter")
    secy_temp2freq = lambda x: temp2freq(x, "freq")
    secy_freq2temp = lambda x: freq2temp(x)

    if not freq_input:
        secx_fcts = (secx_pres2freq, secx_freq2pres)
        secy_fcts = (secy_temp2freq, secy_freq2temp)
        xlab = "Depth (m)"
        ylab = "Temp (°C)"
        secxlab = "Pres Sensor Freq (Hz)"
        secylab = "Temp Sensor Freq (Hz)"
    else:
        secx_fcts = (secx_freq2pres, secx_pres2freq)
        secy_fcts = (secy_freq2temp, secy_temp2freq)
        secxlab = "Depth (m)"
        secylab = "Temp (°C)"
        xlab = "Pres Sensor Freq (Hz)"
        ylab = "Temp Sensor Freq (Hz)"

    if not relative_delta:
        ResVals = ResVals * 10 ** 3
        title = "Error in millimeter max={:.4f}mm".format(ResVals.max())
        format_colorbar = "%.4f"
    else:
        ResVals = ResVals * 10 ** 6
        title = "PPM max={:.4f}".format(ResVals.max())
        format_colorbar = "%.4f"

    ### change values of ResVal

    #################### INITIALISATION ##########################

    fig, ax = plt.subplots()
    ax.ticklabel_format(useOffset=False)
    cm = plt.cm.get_cmap("viridis")
    fig.subplots_adjust(right=0.75, top=0.85)
    im1 = ax.contourf(PresVals, TempVals, ResVals, cmap=cm, levels=200)
    cbar_ax1 = fig.add_axes([0.88, 0.15, 0.04, 0.7])
    cbar1 = fig.colorbar(im1, cax=cbar_ax1, format=format_colorbar)

    secx = ax.secondary_xaxis("top", functions=secx_fcts)
    secy = ax.secondary_yaxis("right", functions=secy_fcts)
    secx.ticklabel_format(useOffset=False)
    secy.ticklabel_format(useOffset=False)

    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)
    secx.set_xlabel(secxlab)
    secy.set_ylabel(secylab)
    ax.set_title(title)

    return fig, ax


#   __                  _   _                                                             _
#  / _|                | | (_)                                                           | |
# | |_ _   _ _ __   ___| |_ _  ___  _ __     __ _ _ __ __ ___   _____ _   _  __ _ _ __ __| |
# |  _| | | | '_ \ / __| __| |/ _ \| '_ \   / _` | '__/ _` \ \ / / _ \ | | |/ _` | '__/ _` |
# | | | |_| | | | | (__| |_| | (_) | | | | | (_| | | | (_| |\ v /  __/ |_| | (_| | | | (_| |
# |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|  \__, |_|  \__,_| \_/ \___|\__, |\__,_|_|  \__,_|
#                                            __/ |                     __/ |
#                                           |___/                     |___/


def freq2pres_old(
        f, U, F0, C1=-22682.65, C2=-1143.743, C3=70903.62, D1=0.040903, D2=0.0
):
    """
    freqence capteur pression > pression
    OLD version
    """

    C = C1 + C2 * U + C3 * U ** 2
    D = D1 + D2 * U

    T = (1 / f) * 10 ** 6
    T0 = (1 / F0) * 10 ** 6

    P = C * (1 - (T0 ** 2) / (T ** 2)) * (1 - D * (1 - (T0 ** 2) / (T ** 2)))
    return P


def temp2freq_old(T, out="freq"):
    """
    temperature > frequence du capteur de temperature (output == "freq")
    OU  coef U (output == "U") OU  periode du capteur U (output == "tau")
    """

    #### Handle T as an interable (array, list....)
    if utils.is_iterable(T):
        return np.array([temp2freq(t, out) for t in T])
    #### Handle T as a scalar
    else:
        U0 = 5.766353
        Y1 = -4025.183
        Y2 = -11970.76
        Y3 = 0.0
        U = np.roots([Y3, Y2, Y1, -T]) + U0
        # X = (1/f) * 10**6 # (1/30000) * 10**6
        # U = X - U0
        # Temp = Y1 * U + Y2 * U**2 + Y3 * U**3
        if out == "freq":
            return [
                10 ** 6 / u for u in U if ((10 ** 6 / u < 176000) and (10 ** 6 / u > 168000))
            ][0]
        elif out == "tau":
            return [
                10 ** 6 / u for u in U if ((10 ** 6 / u < 176000) and (10 ** 6 / u > 168000))
            ][0] ** -1
        elif out == "U":
            return [
                u - U0 for u in U if ((u > 10 ** 6 / 176000) and (u < 10 ** 6 / 168000))
            ][0]
        else:
            return None


def temp2freq_legacy(T, U0, Y, output="freq"):
    # U0 =  5.766353
    # Y1 = -4025.183
    # Y2 = -11970.76
    # Y3 =  0.
    U = np.roots([Y[2], Y[1], Y[0], -T]) + U0
    # X = (1/f) * 10**6 # (1/30000) * 10**6
    # U = X - U0
    # Temp = Y1 * U + Y2 * U**2 + Y3 * U**3
    if output == "freq":
        return [
            10 ** 6 / u for u in U if ((10 ** 6 / u < 176000) and (10 ** 6 / u > 168000))
        ][0]

    else:
        return [u - U0 for u in U if ((u > 10 ** 6 / 176000) and (u < 10 ** 6 / 168000))][0]



"""
Created on Thu Aug 13 15:11:12 2020

@author: psakic

This module geodezyx.externlib import easily all the useful extrern Python's packages.

it can be imported directly with:
from geodezyx.externlib import *  

!!!! THE EXTERN LIB IS JUST A LAZY TRICK FOR THE DEVELOPPER !!!!
!!!! TO IMPORT ALL THE USEFUL PACKAGES IN ONE LINE !!!!
!!!! IT SHOULD NOT BE USED OPERATIONNALY !!!!
!!!! IT IS NOT A GOOD PRACTICE TO IMPORT ALL THE PACKAGES IN ONE LINE !!!!

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""


from .externlib import *

# #### IMPORT EXTERNAL MODULES
import bisect
from collections import defaultdict
from collections import Counter
import copy
import datetime as dt
import fnmatch
from ftplib import FTP
import glob
import inspect
import itertools
import linecache
import math
import matplotlib
import matplotlib.pyplot as plt
import multiprocessing as mp
from natsort import natsorted, ns
import numpy as np
import numpy as npaa
import operator
import os 
from os import remove, close
import pandas as df
import pandas as pd
import pickle
import scipy
from scipy.signal import butter, lfilter, freqz
import shutil
import string
import struct
import sys
import subprocess
import re
import tempfile
from tempfile import mkstemp
import tabulate
import time
import urllib
import uuid




from .conv_angle import *
from .conv_coords import *
from .conv_constant import *
from .conv_geometric import *
from .conv_proj_lambert import *
from .conv_proj_utm import *
from .conv_rinex import *
from .conv_rotation_matrices import *
from .conv_time import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 26/02/2025 11:06:06

@author: psakic
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import os
import time
import numpy as np
import pandas as pd
import yaml

#### geodeZYX modules
from geodezyx import conv
from geodezyx import files_rw
from geodezyx import operational
from geodezyx import utils

#### Import geodezyx GINS submodules
import geodezyx.operational.gins_runner.gins_common as gynscmn
import geodezyx.operational.gins_runner.gins_prairie as gynspra
import geodezyx.operational.gins_runner.legacy.gins_orbclk_convert as gynsorbcvt

#### Import the logger
import logging

log = logging.getLogger("geodezyx")


def gen_dirs_rnxs(
        rnx_paths_inp,
        director_generik_path,
        director_name_prefix,
        out_director_folder=None,
        temp_data_folder=None,
        stations_file=None,
        oceanload_file=None,
        options_prairie_file=None,
        auto_stations_file=False,
        auto_oceanload=False,
        perso_orbclk=False,
        ac="igs",
        repro=2,
        auto_interval=True,
        out_coords="NULL",
        prairie=False,
        prairie_kwargs={"with_historik": 1, "with_wsb": 1},
        force=False,
        verbose=True,
        sites_id9=None,
        add_tropo_sol=True,
):
    """
    Generate directors from RINEX files.

    This function can manage both a single RINEX file or a list of RINEX files.
    It produces a director from a RINEX file and writes it to a specified location.

    If the RINEX is not in a gins style folder, it will be copied in the ``temp_data_folder``

    Parameters
    ----------
    rnx_paths_inp : str or list
        Path to a RINEX file or a list of RINEX file paths.
    director_generik_path : str
        Path to the generic director file.
    director_name_prefix : str
        Prefix for the director name.
    out_director_folder : str, optional
        Output folder for the director. Defaults to None.
        is not specified, output director will be
        created in the ``../gin/data/directeur`` folder
    temp_data_folder : str, optional
        Temporary data folder. Defaults to None.
        is not specified RINEXs will be copied in
        an ad hoc folder ``../gin/TMP_GYNS``
    stations_file : str, optional
        Path to the stations file. Defaults to None.
    oceanload_file : str, optional
        Path to the ocean load file. Defaults to None.
    options_prairie_file : str, optional
        Path to the prairie options file. Defaults to None.
    auto_stations_file: bool, optional
    auto_oceanload: bool, optional
        Automatically create station and ocean loading files.
        create automatically a station file and a ocean loading file
        with the rinex header
        auto mode is prioritary upon the manu mode
        so, if a path for stat file or ocload file is specified and
        auto_staocl is on, it will be the automatic stat/ocload
        which will be used
        Defaults to False.
    perso_orbclk : bool, optional
        Download and use specific orbits/clocks. Defaults to False.
    ac : str, optional
        Analysis center for perso orbits/clocks. Defaults to "igs".
    repro : int, optional
        Reprocessing version for perso orbits/clocks. Defaults to 2.
    auto_interval : bool, optional
        Automatically find the interval in RINEX and apply it to the director.
         Defaults to False.
    out_coords : str, optional
        Output coordinates type ('XYZ' or 'FLH'/'PLH'). Defaults to "NULL".
    prairie : bool, optional
        Run prairie externally. Defaults to False.
        if True, prairie_kwargs control the arguments of the function
        prairie_manual (cf above)
    prairie_kwargs : dict, optional
        Arguments for the prairie_manual function.
        Defaults to {"with_historik": 1, "with_wsb": 1}.
    force : bool, optional
        Force the generation of the director even if a solution already exists.
        Defaults to False.
    verbose : bool
        verbose messages. Default is True.
    sites_id9 : iterable, optional
        A list, Pandas series... containing the site_id9 information.
        This is used to extract the site_id9 if the RINEX name is ambiguous.
    add_tropo_sol : bool, optional
        Add tropospheric solution keys to the director and
        then results in the listing and solution. Defaults to True.

    Returns
    -------
    str or list
        Path of the produced director as a string, or a list of paths if multiple RINEX files are processed.
    """

    # Multi or Single Mode ?
    if type(rnx_paths_inp) is list:
        multimode = True
        rnx_path_lis = rnx_paths_inp
        log.info("******** DIRECTORS GENRATION ********")
    elif type(rnx_paths_inp) is str:
        multimode = False
        rnx_path_lis = [rnx_paths_inp]
    else:
        log.error("check the rinex_paths_in !!!")
        return None

    gin_path = gynscmn.get_gin_path()  # must remain not extended
    rnx_path_lis = list(sorted(rnx_path_lis))
    n_rnxs = len(rnx_path_lis)
    director_output_path_lis = []
    failed_rinex_date_lis = []
    temp_data_folder_lis = []

    dir_out_path = None
    tmp_fld_use = None

    def _fail_rnx(rnx_path_inp, rnx_dt_inp, message):
        log.error(message + ", skip %s", rnx_path_inp)
        failed_rinex_date_lis.append(rnx_dt_inp)
        bool_continue = False
        if multimode:
            bool_continue = True
        return bool_continue
        # else:
        #    raise Exception

    for i, rnx_path_ori in enumerate(rnx_path_lis):
        if multimode:
            log.info(f"======== {i + 1} / {n_rnxs} ======== ")
        log.info(" *** Generate directeur for: %s ", os.path.basename(rnx_path_ori))
        rnx_name = os.path.basename(rnx_path_ori)
        rnx_dt = conv.rinexname2dt(rnx_name)
        sites_id9_series = pd.Series(sites_id9)
        siteid9, siteid4_up, siteid4_lo = _dir_rnx_site_id(rnx_name, sites_id9_series)

        coord_prefix = (
            f"_{out_coords.lower()}" if out_coords.upper() in ("FLH", "XYZ") else ""
        )

        ac_suffix = f"_{ac}{repro}" if perso_orbclk else ""

        dir_name = utils.join_improved(
            "_",
            director_name_prefix,
            siteid9,
            str(rnx_dt.year),
            conv.dt2doy(rnx_dt),
            str(conv.dt2jjul_cnes(rnx_dt)),
            # freq_rnx_str,
            coord_prefix,
            # ses,
            ac_suffix,
        )

        sols_matching = gynscmn.check_solution(dir_name)

        if len(sols_matching) > 0 and not force:
            _fail_rnx(rnx_path_ori, rnx_dt, "solution already exists")
            continue

        # be sure there is a TEMP DATA folder
        if not temp_data_folder:
            temp_data_folder = gynscmn.get_temp_data_gins_path()
        if not os.path.exists(temp_data_folder):
            os.makedirs(temp_data_folder)

        # go for a sub TEMP folder
        tmp_fld_use = os.path.join(
            temp_data_folder, utils.get_timestamp() + "_" + dir_name
        )
        os.makedirs(tmp_fld_use)

        # be sure the RINEX is in gins folder ...
        rnx_path = gynscmn.bring_to_gin(rnx_path_ori, tmp_fld_use, verbose=verbose)

        # check if the RINEX is compressed ... if not crz2rnx !
        if operational.check_if_compressed_rinex(rnx_path):
            crinex_path = rnx_path
            rnx_path = operational.crz2rnx(crinex_path, tmp_fld_use, verbose=verbose)
            if not os.path.isfile(rnx_path):
                bool_cntu = _fail_rnx(rnx_path, rnx_dt, "CRZ2RNX failed")
                if bool_cntu:
                    continue

        # prairie extern
        pra_file_path = None
        if prairie:
            log.info("run prairie externally")
            pra_file_path = gynspra.prairie_manual(
                rnx_path, tmp_fld_use, **prairie_kwargs
            )

            if type(pra_file_path) is list and len(pra_file_path) == 0:
                bool_cntu = _fail_rnx(rnx_path, rnx_dt, "manual prairie failed")
                if bool_cntu:
                    continue

        if not out_director_folder:
            out_director_folder = os.path.join(gin_path, "gin", "data", "directeur")

        # date

        srt_epo, end_epo, freq_rnx = None, None, None

        try:
            srt_epo, end_epo, freq_rnx = operational.rinex_start_end(rnx_path, True, verbose=verbose)
        except Exception:
            log.warning("Unable to read RINEX, fallback to filename to get its start/end")
            srt_epo = rnx_dt
            end_epo = srt_epo + dt.timedelta(seconds=86360)  # 1 day
            freq_rnx = 30

        if not srt_epo and not end_epo and not freq_rnx:
            srt_epo, end_epo, freq_rnx = None, None, None
            bool_cntu = _fail_rnx(rnx_path, rnx_dt, "get RINEX start/end failed")
            if bool_cntu:
                continue

        dir_out_fname = dir_name + ".yml"
        dir_out_path = os.path.join(out_director_folder, dir_out_fname)
        director_generik_file = open(director_generik_path)
        dir_dic = yaml.load(director_generik_file, Loader=yaml.FullLoader)

        # ADDING RINEX NAME & DATES INTO THE DIR
        # name
        if not prairie:
            rnx_path_gstyl = gynscmn.make_path_ginsstyle(rnx_path)
            dir_dic["observation"]["interobject_data"][1]["file"] = rnx_path_gstyl
        else:
            pra_path_dir_gstyl = gynscmn.make_path_ginsstyle(pra_file_path)
            dir_dic["observation"]["interobject_data"][1]["file"] = pra_path_dir_gstyl

        # date
        strt_day, strt_sec = conv.dt2jjul_cnes(srt_epo, False)
        end_day, end_sec = conv.dt2jjul_cnes(end_epo, False)

        # SPOTGINS compatibility for the full day
        # one more sec will be +/- below
        strt_sec = 19
        end_sec =  86389.0

        dir_dic["date"]["arc_start"][0] = strt_day
        dir_dic["date"]["arc_stop"][0] = end_day
        if "initial_state_vector_date" in list(dir_dic["date"].keys()):
            dir_dic["date"]["initial_state_vector_date"][0] = strt_day

        dir_dic["date"]["arc_start"][1] = (
                strt_sec - 1
        )  # -1 to make it SPOTGINS compatible
        dir_dic["date"]["arc_stop"][1] = (
                end_sec + 1
        )  # +1 to make it SPOTGINS compatible
        if "initial_state_vector_date" in list(dir_dic["date"].keys()):
            dir_dic["date"]["initial_state_vector_date"][1] = strt_sec

        # interval
        if auto_interval:  # dir is modified only if freq_rnx >= 0
            dir_dic = _dir_auto_intrvl(dir_dic, freq_rnx)

        # output coords
        if out_coords.upper() in ("FLH", "PLH"):
            dir_dic["parameter"]["adjustment_parameters"]["stations"][
                "coordinates"
            ] = "latitude_longitude_height"
        elif out_coords.upper() == "XYZ":
            dir_dic["parameter"]["adjustment_parameters"]["stations"][
                "coordinates"
            ] = "cartesian_xyz"

        # ============== STATION FILE & OCEANLOAD FILE ==============
        # station file and oceanload file , auto or manu mode
        # auto mode is prioritary upon the manu mode

        if not auto_stations_file and not stations_file:
            log.warning("auto_stations_file is off and no stations_file specified !!!")

        if not auto_oceanload and not oceanload_file:
            log.warning("auto_oceanload is off and no oceanload_file specified !!!")

        if auto_stations_file:
            stations_file = _dir_auto_stfi(siteid4_lo, rnx_dt, rnx_path, tmp_fld_use)

        if auto_oceanload:
            oceanload_file = _dir_auto_oclo(
                siteid4_lo, rnx_dt, tmp_fld_use, stations_file
            )

        if stations_file:
            stfi_ingin = gynscmn.bring_to_gin(str(stations_file), tmp_fld_use, verbose=verbose)
            dir_dic["object"]["station"]["station_coordinates"] = (
                gynscmn.make_path_ginsstyle(stfi_ingin)
            )
        if oceanload_file:
            oclo_ingin = gynscmn.bring_to_gin(str(oceanload_file), tmp_fld_use, verbose=verbose)
            dir_dic["object"]["station"]["ocean_tide_loading"] = (
                gynscmn.make_path_ginsstyle(oclo_ingin)
            )

        # ============== PRAIRIE OPTIONS ==============
        # NB : prairie options path does not have to be in gin style
        # but in its absolute path
        if options_prairie_file:
            dir_dic["model"]["environment"]["gnss_preprocessing_options"] = (
                os.path.realpath(options_prairie_file)
            )

        # ============== ORBITS/CLOCKS ================
        if perso_orbclk:
            orbpath, horpath = gynsorbcvt.download_convert_2_gins_orb_clk(
                rnx_dt, tmp_fld_use, ac=ac, repro=repro
            )
        else:
            orbpath, horpath = _dir_regular_orbclk(rnx_dt)

        # Exception case where no sp3/clk was found
        if orbpath is None or horpath is None:
            bool_cntu = _fail_rnx(rnx_path, rnx_dt, "orbits/clocks download/conversion")
            if bool_cntu:
                continue

        if perso_orbclk:
            orbpath = gynscmn.make_path_ginsstyle(orbpath)
            horpath = gynscmn.make_path_ginsstyle(horpath)

        dir_dic["model"]["environment"]["gnss_clock"] = horpath
        dir_dic["observation"]["interobject_data"][0]["file"] = orbpath

        # Security CHECKs
        stfi_path_gstyl = dir_dic["object"]["station"]["station_coordinates"]
        stfi_path_full = str(os.path.join(gin_path, stfi_path_gstyl[6:]))

        oclo_path_gstyl = dir_dic["object"]["station"]["ocean_tide_loading"]
        oclo_path_full = str(os.path.join(gin_path, oclo_path_gstyl[6:]))

        gynscmn.check_site_stfl(siteid4_up, stfi_path_full)

        domes = gynscmn.find_domes_in_stfl(siteid4_up, stfi_path_full)
        if verbose:
            log.info("DOMES: %s", domes)
        gynscmn.chek_domes_oclo(domes[0], oclo_path_full)

        # ========= CUSTOM USER EXTENSION (tropo, high freq...) =============
        dir_dic = _dir_userext(dir_dic, siteid9, add_tropo_sol)

        # WRITING THE NEW DIRECTOR
        if verbose:
            log.info("writing: %s", dir_out_path)
        with open(dir_out_path, "w+") as outfile:
            outfile.write(yaml.dump(dir_dic, default_flow_style=False))

        # correcting a bug : GINS YAML interpreter can't manage false or true ...
        utils.replace(dir_out_path, "false", "no")
        utils.replace(dir_out_path, "true", "yes")

        director_output_path_lis.append(dir_out_path)
        temp_data_folder_lis.append(tmp_fld_use)

    if multimode:
        if len(failed_rinex_date_lis) > 0:
            print("INFO : ", len(failed_rinex_date_lis), "/", n_rnxs, " RINEXs failed")
            print("       ", [str(d) for d in failed_rinex_date_lis])
        return director_output_path_lis, temp_data_folder_lis
    else:
        return dir_out_path, tmp_fld_use


def _dir_regular_orbclk(dt_rinex_inp):
    """
    Get the regular orbits and clocks path for a given date
    """

    if dt_rinex_inp <= dt.datetime.now() - dt.timedelta(days=14):
        prod_id = "G20"
    else:
        prod_id = "G20"

    ### ALL THE OTHER ORBITS/CLOCKS ARE NOW OBSOLETE (2025)

    horpath = os.path.join("horloges", prod_id, "defaut")
    orbpath = os.path.join("orbites", prod_id, "defaut")

    return orbpath, horpath


def _dir_auto_intrvl(dir_dic, freq_rnx):
    """
    Automatically set the interval in the director
    """
    dir_dic["observation"]["removal"]["simulation_stepsize"] = freq_rnx
    # check if its not a static dir
    if (
            dir_dic["parameter"]["adjustment_parameters"]["stations"][
                "adjustment_frequency"
            ][-1]
            != 0
    ):
        dir_dic["parameter"]["adjustment_parameters"]["stations"][
            "adjustment_frequency"
        ] = [0, 0, 0, freq_rnx]
    log.info(
        "auto_interval, simulation_stepsize : %s",
        dir_dic["observation"]["removal"]["simulation_stepsize"],
    )

    return dir_dic


def _dir_auto_stfi(stat_lower, dt_rinex, rinex_path, temp_data_folder):
    """
    Generate an ad hoc station file from a RINEX file
    """
    log.info("* Automatic station file generation")
    randid = "id" + str(np.random.randint(10000, 99999))
    stat_fname = utils.join_improved(
        "_",
        stat_lower,
        str(dt_rinex.year),
        str(conv.dt2doy(dt_rinex)),
        randid,
        ".stat",
    )

    stfi_path_out = os.path.join(temp_data_folder, stat_fname)
    files_rw.write_station_file_gins_from_rinex(rinex_path, stfi_path_out)
    bool_statfil = os.path.isfile(stfi_path_out)
    if not os.path.isfile(stfi_path_out):
        log.error("no station file genrated !!!")
        stfi_path_out = None

    return stfi_path_out


def _dir_auto_oclo(stat_lower, dt_rinex, temp_data_folder, stat_path_inp):
    """
    Generate an ad hoc ocean loading file from a station file
    """
    log.info("* Automatic stat and ocean loading file generation")
    randid = "id" + str(np.random.randint(10000, 99999))

    oclo_fname = utils.join_improved(
        "_",
        stat_lower,
        str(dt_rinex.year),
        str(conv.dt2doy(dt_rinex)),
        randid,
        ".oclo",
    )

    oclo_path_out = os.path.join(temp_data_folder, oclo_fname)
    gynscmn.write_oclo_file(stat_path_inp, oclo_path_out)

    bool_oclofil = os.path.isfile(oclo_path_out)
    iwhile = 0
    while not bool_oclofil or iwhile < 10:
        if iwhile > 6:
            log.warning("waiting for the ocean loading file for a while %i", iwhile)
        iwhile = iwhile + 1
        time.sleep(0.5)
        bool_oclofil = os.path.isfile(oclo_path_out)

    if not os.path.isfile(oclo_path_out):
        log.error("no oceload file genrated !!!")
        oclo_path_out = None

    return oclo_path_out


def _dir_rnx_site_id(rnx_name, sites_id9_series):
    """
    Extract the site_id9 from a Series (from e.g. a SPOTGINS master file)
    based on the RINEX name
    """
    if conv.rinex_regex_search_tester(
            rnx_name, short_name=False, long_name=True
    ):  ### RINEX3
        site_id9 = rnx_name[0:9].upper()
    else:  ### RINEX2
        site_id4 = rnx_name[0:4].upper()
        site_id9 = site_id4 + "00XXX"
        if type(sites_id9_series) is pd.Series:
            ser_bool = sites_id9_series.str[:4].str.match(site_id4)
            if ser_bool.any():
                site_id9 = sites_id9_series.loc[ser_bool].values[0]
            if ser_bool.sum() > 1:
                log.warning("more than one site_id9 found for %s: %s", site_id4,
                            sites_id9_series.loc[ser_bool].values)
                log.warning("taking the first one : %s", site_id9)

    site_id4_upper = site_id9[0:4]
    site_id4_lower = site_id4_upper.lower()

    return site_id9, site_id4_upper, site_id4_lower


def _dir_userext(dir_dic, site_id9, add_tropo_sol=True):
    """
    Customize the user_extension fields with the site_id4_upper
    """
    uext = "user_extension"
    uadd = "userext_addition"

    if uext in dir_dic:
        if add_tropo_sol:
            dir_dic[uext][uadd].append("ADD_TROPO_IN_SOLUTION")
            dir_dic[uext][uadd].append("GPS__AFF_SINEX")

        # GPS__HAUTE_FREQ STAT is not a key/value couple, but is considered a single string
        # we must search this string and concatenate it with the right site code
        if uadd in dir_dic[uext]:
            is_gpshf = ["GPS__HAUTE_FREQ" in e for e in dir_dic[uext][uadd]]
            if np.any(is_gpshf):
                idx_gpshf = is_gpshf.index(True)
                dir_dic[uext][uadd][idx_gpshf] = "GPS__HAUTE_FREQ " + site_id9.upper()

    return dir_dic


def _dir_auto_staocl(stat_lower, dt_rinex, rinex_path, temp_data_folder):
    ##### DEPRECATED, NOW SEPARATED !!!
    log.info("* Automatic stat and ocload files generation")
    randid = "id" + str(np.random.randint(10000, 99999))
    stat_fname = utils.join_improved(
        "_",
        stat_lower,
        str(dt_rinex.year),
        str(conv.dt2doy(dt_rinex)),
        randid,
        ".stat",
    )

    stat_path_out = os.path.join(temp_data_folder, stat_fname)
    files_rw.write_station_file_gins_from_rinex(rinex_path, stat_path_out)
    oclo_fname = utils.join_improved(
        "_",
        stat_lower,
        str(dt_rinex.year),
        str(conv.dt2doy(dt_rinex)),
        randid,
        ".oclo",
    )

    oclo_path_out = os.path.join(temp_data_folder, oclo_fname)
    gynscmn.write_oclo_file(stat_path_out, oclo_path_out)

    bool_statfil = os.path.isfile(stat_path_out)
    bool_oclofil = os.path.isfile(oclo_path_out)
    iwhile = 0
    while not (bool_statfil and bool_oclofil) or iwhile < 10:
        if iwhile > 6:
            log.warning("waiting for the ocean loading file for a while %i", iwhile)
        iwhile = iwhile + 1
        time.sleep(0.5)
        bool_statfil = os.path.isfile(stat_path_out)
        bool_oclofil = os.path.isfile(oclo_path_out)

    if not os.path.isfile(stat_path_out):
        log.error("no station file genrated !!!")
        stat_path_out = None

    if not os.path.isfile(oclo_path_out):
        log.error("no oceload file genrated !!!")
        oclo_path_out = None

    return stat_path_out, oclo_path_out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.utils contains functions for operations
related to Python's list manipulations.

it can be imported directly with:
from geodezyx import utils

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""


########## BEGIN IMPORT ##########
#### Import the logger
import logging
import re
import os

import numpy as np

#### geodeZYX modules
log = logging.getLogger("geodezyx")


##########  END IMPORT  ##########


def is_listoflist(inp):
    """
    check if inp is a list of list [[...] , [...] , ... ,[...]]
    """
    # On fait l'hypothèse que c'est un iterable qui va dedans de toute façon
    return any(isinstance(el, (list, np.ndarray)) for el in inp)


def shrink_listoflist(lin):
    """
    if lin (list in) is a list of list and contains only one element
    then returns the sublist, e.g. : [[a,b,c]] => [a,b,c]
    """
    if is_listoflist(lin) and len(lin) == 1:
        return lin[0]
    else:
        return lin


def uniqify_list(seq, idfun=None):
    """
    order preserving uniq function
    based on
    https://www.peterbe.com/plog/uniqifiers-benchmark
    """
    if idfun is None:

        def idfun(x):
            return x

    seen = {}
    result = []
    for item in seq:
        marker = idfun(item)
        # in old Python versions:
        # if seen.has_key(marker)
        # but in new ones:
        if marker in seen:
            continue
        seen[marker] = 1
        result.append(item)
    return result


def uniqify_list_of_lists(L):
    """
    source
    http://stackoverflow.com/questions/3724551/python-uniqueness-for-list-of-lists
    """
    return [list(x) for x in set(tuple(x) for x in L)]


def find_common_elts(*lists):
    """
    Find common elements in different lists
    """
    sett = set(lists[0])

    for l in lists[1:]:
        sett = sett.intersection(l)

    return np.array(sorted(list(sett)))


def uniq_and_sort(L, natural_sort=True):
    """
    In a list, remove duplicates and sort the list
    """
    import natsort
    if natural_sort:
        return natsort.natsorted(list(set(L)))
    else:
        return sorted(list(set(L)))


def df_sel_val_in_col(DF, col_name, col_val):
    """
    Return a selected value of a column in a DataFrame
    i.e.
    return DF[DF[col_name] == col_val]
    """
    return DF[DF[col_name] == col_val]


def uniquetol(A, tol):
    """
    tol : absolute tolerance
    source
    http://stackoverflow.com/questions/37847053/uniquify-an-array-list-with-a-tolerance-in-python-uniquetol-equivalent
    """
    A = np.array(A)
    return A[~(np.triu(np.abs(A[:, None] - A) <= tol, 1)).any(0)]


def uniquetol2(A, tol=10**-6):
    """
    This one is faster
    https://stackoverflow.com/questions/5426908/find-unique-elements-of-floating-point-array-in-numpy-with-comparison-using-a-d
    """
    return np.unique(A.round(decimals=4))


def groups_near_central_values(A, tol, B=None):
    """
    beta ... bug if tol is bad
    tol : absolute tolerance
    170514 : B is an annex list
    """
    A = np.array(A)

    Auniq = uniquetol(A, tol)
    Grouplis = []

    if B is None:
        Bbool = False
    else:
        Bbool = True

    if Bbool:
        GrouplisB = []

    for iauniq, auniq in enumerate(Auniq):
        Group = []
        Grouplis.append(Group)

        if Bbool:
            GroupB = []
            GrouplisB.append(GroupB)

        for ia, a in enumerate(A):
            count = 0
            if np.abs(a - auniq) <= tol:
                Group.append(a)
                count += 1

                if Bbool:
                    GroupB.append(B[ia])

        if count > 1:
            log.warning("%s in %s groups", a, count)

    if Bbool:
        return Grouplis, GrouplisB
    else:
        return Grouplis


def uniq_set_list(setlis, frozen=True):
    """uniqify a list of sets"""
    if not frozen:
        return [set(e) for e in list(set([frozenset(e) for e in setlis]))]
    else:
        return [frozenset(e) for e in list(set([frozenset(e) for e in setlis]))]


def sort_binom_list(X, Y, array_out=False):
    """
    sort Y according to X
    and sort X
    """
    if len(X) != len(Y):
        log.warning("len(X) != len(Y) !!!")

    Ynew = [y for (x, y) in sorted(zip(X, Y))]
    Xnew = sorted(X)
    if not array_out:
        return Xnew, Ynew
    else:
        return np.array(Xnew), np.array(Ynew)


def sort_multinom_list(X, *Y):
    """
    sort Y according to X
    and sort X
    """
    Ynew_stk = []
    for YY in Y:
        Ynew = [y for (x, y) in sorted(zip(X, YY))]
        Ynew_stk.append(Ynew)
    Xnew = sorted(X)
    fintup = (np.array(Xnew),) + tuple(Ynew_stk)
    return fintup


def sort_basename(file_paths):
    """
    Sort a list of file paths by their basenames.

    Parameters
    ----------
    file_paths : list
        List of file paths to be sorted.

    Returns
    -------
    list
        Sorted list of file paths by their basenames.
    """
    return list(sorted(file_paths, key=os.path.basename))


def most_common(lst):
    """
    http://stackoverflow.com/questions/1518522/python-most-common-element-in-a-list
    """
    if type(lst) is not list:
        lst = list(lst)
    return max(set(lst), key=lst.count)


def median_improved(L):
    """
    manage the case where len(L) is even
    in this case, doesn't return the mean of the 2 medians, but the nearest value
    """
    L = np.array(L)
    if len(L) % 2 == 0:
        return find_nearest(L, np.median(L))[0]
    else:
        return np.median(L)


def trio_lists_2_tab(Xlis, Ylis, Vlis):
    """
    From a trio of lists
    Xlis => parameter you want in columns
    Ylis => parameter you want in rows
    Vlis => data(X,Y)

    make a tab compatible with mabular module
    tabulate(Finalis, headers="firstrow")

    NB : very dirty research of the elements ...
    """

    Xlis_uniq = sorted(np.unique(Xlis))
    Ylis_uniq = sorted(np.unique(Ylis))

    Finalis = []
    Finalis.append(Xlis_uniq)

    for y in Ylis_uniq:
        curlis = [y]
        Finalis.append(curlis)
        for x in Xlis_uniq:
            for xx, yy, vv in zip(Xlis, Ylis, Vlis):
                if xx == x and yy == y:
                    curlis.append(vv)
    return Finalis


def minmax(L):
    return np.min(L), np.max(L)


def middle(Lin):
    Lout = []
    for i in range(len(Lin) - 1):
        Lout.append((Lin[i] + Lin[i + 1]) / 2)
    return Lout


def second_smallest(numbers):
    m1, m2 = float("inf"), float("inf")
    for x in numbers:
        if x <= m1:
            m1, m2 = x, m1
        elif x < m2:
            m2 = x
    return m2


def find_index_multi_occurences(L, elt):
    return [i for i, x in enumerate(L) if x == elt]


def find_surrounding(L, v):
    """
    find the surrounding values

    Parameters
    ----------
    L : Iterable
        Input list/array.

    Returns
    -------
    surrounding_values : tuple
        surounding values.
    surrounding_index : tuple
        surounding indexes.


    """
    a = np.array(L)
    b = v

    surrounding_values = np.sort(
        [b + i for i in sorted(np.subtract(a, b), key=lambda x: abs(x))[:2]]
    )
    surrounding_index = (
        np.where(surrounding_values[0] == a)[0][0],
        np.where(surrounding_values[1] == a)[0][0],
    )

    return tuple(surrounding_values), surrounding_index


##################
### LIST IT FCTs
##################


def chunkIt(seq, num):
    """make num sublists of a list"""
    # http://stackoverflow.com/questions/2130016/splitting-a-list-of-arbitrary-size-into-only-roughly-n-equal-parts
    avg = len(seq) / float(num)
    out = []
    last = 0.0
    while last < len(seq):
        out.append(seq[int(last) : int(last + avg)])
        last += avg
    return out


def sliceIt(seq, num):
    """make sublist of num elts of a list"""
    # http://stackoverflow.com/questions/4501636/creating-sublists
    return [seq[i : i + num] for i in range(0, len(seq), num)]


def sublistsIt(seq, lenofsublis_lis, output_array=False):
    """
    make sublists of seq , accoding to len of the sublist in lenofsublis_lis
    ex lenofsublis_lis = [2,3,4,2]

    if output_array : output list of array , else list of list

    (fct perso)
    """

    if np.sum(lenofsublis_lis) != len(seq):
        raise Exception("sublistsIt : sum(lenofsublis_lis) != len(seq) ")
    seq = list(seq)
    sublis_lis = []
    start = 0
    for l in lenofsublis_lis:
        end = start + l
        sublis = seq[start:end]
        sublis_lis.append(sublis)
        start = end
    if output_array:
        return [np.array(e) for e in sublis_lis]
    else:
        return sublis_lis


def identical_consecutive_eltsIt(Lin):
    Lout_big = []
    Linter = [Lin[0]]
    Lout_big.append(Linter)

    for e in Lin[1:]:
        if e == Linter[-1]:
            Linter.append(e)
        else:
            Linter = [e]
            Lout_big.append(Linter)

    return Lout_big


def find_nearest(listin, value):
    """
    Returns:
        value of the nearest , index of the nearest
    """
    array = np.array(listin)
    idx = (np.abs(array - value)).argmin()

    return listin[idx], idx


def find_interval_bound(listin, val, outindexes=True):
    """
    trouve les bornes d'un intervalle
    (listin est supposé triée)
    """
    import bisect
    i = bisect.bisect(listin, val)
    if outindexes:
        return i - 1, i
    else:
        return listin[i - 1], listin[i]


def occurence(L, tolerence=None, pretty_output=False):
    """
    Input :
        tolerence : tolerence to find close elements of L
                    if no tolerence is given then a set() is used
    Returns :
        if pretty_output = False :
            return a list of 2-tuples : (element of the list, number of occurence of this element in the list)
        if pretty_output 0 True :

    Nota : pretty_output is implemented because the first mode is not really useful (180612)
           the equal test is also replaced by is close
    """
    if tolerence:
        Lset = uniquetol2(L, tol=tolerence)
    else:
        Lset = set(L)

    L = np.array(L)
    outlisoccur = []
    for l in Lset:
        outlisoccur.append((l, np.sum(L == l)))

    if pretty_output:
        Mtmp = np.vstack(outlisoccur)
        Vals = Mtmp[:, 0]
        Occurs = Mtmp[:, 1]
        Occurs, Vals = sort_binom_list(Occurs, Vals)

        output = (Occurs, Vals)
    else:
        output = outlisoccur

    return output


def decimateIt(listin, n):
    """
    n so as np.mod(i,n) == 0
    """
    outlist = []
    for i in range(len(listin)):
        if np.mod(i, n) == 0:
            outlist.append(listin[i])
    return outlist


def consecutive_groupIt(data, only_start_end=False):
    """
    Identify groups of continuous numbers in a list

    Useful for time period with a prealable conversion to MJD

    Source :
    https://stackoverflow.com/questions/2154249/identify-groups-of-continuous-numbers-in-a-list
    """
    from operator import itemgetter
    import itertools

    groups_stk = []
    for k, g in itertools.groupby(enumerate(data), lambda A: A[0] - A[1]):
        groups_stk.append(list(map(itemgetter(1), g)))

    if only_start_end:
        return [(g[0], g[-1]) for g in groups_stk]
    else:
        return groups_stk


def identical_groupIt(data):
    """
    Source :
        https://stackoverflow.com/questions/30293071/python-find-same-values-in-a-list-and-group-together-a-new-list
    """
    import itertools
    return [list(j) for i, j in itertools.groupby(data)]


def get_interval(start, end, delta):
    # after http://stackoverflow.com/questions/10688006/generate-a-list-of-datetimes-between-an-interval-in-python
    outlist = []
    curr = start
    while curr < end:
        outlist.append(curr)
        curr += delta
    return outlist


def duplicates_finder(seq):
    """
    from
    http://stackoverflow.com/questions/9835762/find-and-list-duplicates-in-python-list
    moooeeeep solution
    """
    seen = set()
    seen_add = seen.add
    # adds all elements it doesn't know yet to seen and all other to seen_twice
    seen_twice = set(x for x in seq if x in seen or seen_add(x))
    # turn the set into a list (as requested)
    return list(seen_twice)


def sort_table(table, col):
    """sort a table by multiple columns
    table: a list of lists (or tuple of tuples) where each inner list
           represents a row
    cols: specifying the column numbers to sort by
    """
    refcol = list(table[col])
    outtable = []
    for col in table:
        outtable.append([x for (y, x) in sorted(zip(refcol, col))])
    return outtable


def dicofdic(mat, names):

    # a partir d'un matrice de N x N
    # et d'une liste de N noms
    # Fabrique un dictionnaire 2D dic[nom1][nom2] = M[n1,n2]
    # http://stackoverflow.com/questions/13326042/2d-dictionary-with-multiple-keys-per-value
    import collections
    d2_dict = collections.defaultdict(dict)

    for i in range(mat.shape[0]):
        for j in range(mat.shape[1]):
            d2_dict[names[i]][names[j]] = mat[i, j]

    return d2_dict


def find_regex_in_list(regex, L, only_first_occurence=False, line_number=False):
    Lout = []
    for i, e in enumerate(L):
        if re.search(regex, e):
            if not line_number:
                found = e
            else:
                found = (i, e)

            if only_first_occurence:
                return found
            else:
                Lout.append(found)
    return Lout



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb  6 10:30:57 2024

Filière ING3 - PPMD - Traitement de la mesure de phase

@author: Samuel Nahmani (1,2)
https://www.ipgp.fr/annuaire/nahmani/)
contact : nahmani@ipgp.fr ou samuel.nahmani@ign.fr
(1) Université Paris Cité, Institut de physique du globe de Paris, CNRS, IGN, F-75005 Paris, France.
(2) Univ Gustave Eiffel, ENSG, IGN, F-77455 Marne-la-Vallée, France. 

Version: 1.0
Dépendances: numpy, geodezyx, datetime

"""
#%%

"""
Conversion du code Matlab de Mahooti (2024) en Python
Meysam Mahooti (2024). Klobuchar Ionospheric Delay Model (https://www.mathworks.com/matlabcentral/fileexchange/59530-klobuchar-ionospheric-delay-model), MATLAB Central File Exchange. Retrieved February 6, 2024. 

"""


#%%
import numpy as np
from geodezyx import conv
import datetime as dt


def klobuchar(phi, lambda_, elev, azimuth, tow, alpha, beta):
    """
    Compute ionospheric range correction for GPS L1 frequency using Klobuchar model.
    
    Translation to Python of Meysam Mahooti (2024). Klobuchar Ionospheric Delay Model 
    https://www.mathworks.com/matlabcentral/fileexchange/59530-klobuchar-ionospheric-delay-model
    MATLAB Central File Exchange. Retrieved February 7, 2024.
    
    Parameters
    ----------
    phi : float
        Geodetic latitude of receiver (degrees).
    lambda_ : float
        Geodetic longitude of receiver (degrees).
    elev : float
        Elevation angle of satellite (degrees).
    azimuth : float
        Geodetic azimuth of satellite (degrees).
    tow : float
        Time of Week (seconds).
    alpha : array_like
        The coefficients of a cubic equation representing the amplitude 
        of the vertical delay (4 coefficients - 8 bits each).
    beta : array_like
        The coefficients of a cubic equation representing the period 
        of the model (4 coefficients - 8 bits each).
    
    Returns
    -------
    d_ion1 : float
        Ionospheric slant range correction for the L1 frequency (metres).
    
    References
    ----------
    Klobuchar, J.A., (1996) "Ionospheric Effects on GPS", in
    Parkinson, Spilker (ed), "Global Positioning System Theory and
    Applications, pp.513-514.
    
    ICD-GPS-200, Rev. C, (1997), pp. 125-128
    
    NATO, (1991), "Technical Characteristics of the NAVSTAR GPS",
    pp. A-6-31   -   A-6-33
    """

    # Constants
    c = 2.99792458e8  # Speed of light
    deg2semi = 1 / 180  # Degrees to semicircles
    semi2rad = np.pi  # Semicircles to radians
    deg2rad = np.pi / 180  # Degrees to radians

    # Convert angles to radians or semicircles
    a = azimuth * deg2rad  # Azimuth in radians
    e = elev * deg2semi    # Elevation angle in semicircles

    # Earth Centered angle
    psi = 0.0137 / (e + 0.11) - 0.022
    lat_i = phi * deg2semi + psi * np.cos(a)  # Subionospheric latitude
    lat_i = np.clip(lat_i, -0.416, 0.416)    # Clipping lat_i

    # Subionospheric longitude
    long_i = lambda_ * deg2semi + (psi * np.sin(a) / np.cos(lat_i * semi2rad))

    # Geomagnetic latitude
    lat_m = lat_i + 0.064 * np.cos((long_i - 1.617) * semi2rad)

    # Time
    t = 4.32e4 * long_i + tow
    t = t % 86400  # Seconds of day

    # Slant factor
    s_f = 1 + 16 * (0.53 - e) ** 3

    # Period of model
    per = beta[0] + beta[1] * lat_m + beta[2] * lat_m ** 2 + beta[3] * lat_m ** 3
    per = max(per, 72000)

    # Phase of the model
    x = 2 * np.pi * (t - 50400) / per

    # Amplitude of the model
    amp = alpha[0] + alpha[1] * lat_m + alpha[2] * lat_m ** 2 + alpha[3] * lat_m ** 3
    amp = max(amp, 0)

    # Ionospheric correction
    if np.abs(x) > 1.57:
        d_ion1 = s_f * 5e-9
    else:
        d_ion1 = s_f * (5e-9 + amp * (1 - x ** 2 / 2 + x ** 4 / 24))

    # Ionospheric slant range correction
    d_ion1 = c * d_ion1
    return d_ion1

if __name__ == "__main__":
    
   # Coefficients alpha et beta
    alpha = np.array([0.382e-7, 0.149e-7, -0.179e-6, 0.0])
    beta = np.array([0.143e6, 0.0, -0.328e6, 0.113e6])
    
    # Temps UTC
    year = 2000
    month = 1
    day = 1
    hour = 20
    minute = 45
    second = 0
    
    # Autres paramètres
    elev = 20
    azimuth = 210
    phi = 40
    lambda_ = 260
    
    # Calcul du TOW
    # Exemple d'utilisation
    t = dt.datetime(year, month, day, hour, minute, second)
    week, tow = conv.dt2gpstime(t,secinweek=True)
    print("GPS Week:", week)
    print("GPS Seconds:", tow)
 
    
    # Correction ionosphérique L1
    dIon1 = klobuchar(phi, lambda_, elev, azimuth, tow, alpha, beta)
    print("Ionospheric L1 Correction: ", dIon1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec  1 17:46:14 2023

@author: psakic
"""

#### Import the logger
import logging
import os
import re
from io import StringIO

import numpy as np
import pandas as pd

log = logging.getLogger('geodezyx')

#                       _     _     
#     /\               | |   (_)    
#    /  \   _ __  _   _| |__  _ ___ 
#   / /\ \ | '_ \| | | | '_ \| / __|
#  / ____ \| | | | |_| | |_) | \__ \
# /_/    \_\_| |_|\__,_|_.__/|_|___/
                        

def read_anubis_xtr_sum(xtr_in):
    
    if os.path.basename(xtr_in).split(".")[-1]  != "xtr":
        log.warn("%s is not a XTR file",xtr_in)
    
    L = open(xtr_in).readlines()
    
    def _date_format(linein,end_epoch=False):
        ltmp = list(linein)
        ltmp[18] = "T"
        if end_epoch:
            ltmp[38] = "T"
        return "".join(ltmp)
    
    sum_pattern = re.compile(r"^(#|=)...SUM ")
    
    totsum_stk = []
    gnssum_stk = []
    
    for line in L:
        sum_match = sum_pattern.match(line)
        if sum_match:
            if "TOTSUM" in line:
                totsum_stk.append(line)
            else:
                gnssum_stk.append(line)
                
        if "Header information" in line:
            break
    
    ########### GNSSUM
    ## remove 1st line of GNSSUM
    gnssum_stk.remove(gnssum_stk[0])
    
    for il,l in enumerate(gnssum_stk):
        gnssum_stk[il] = _date_format(l,False)
    
    df_gnssum = pd.read_csv(StringIO("\n".join(gnssum_stk)),
                            delim_whitespace=True,
                            na_values="-")
    
    df_gnssum.rename({"#GNSSUM":"sys",
                      df_gnssum.columns[1]:"Epoch"},axis=1,inplace=True)
    df_gnssum["sys"] = df_gnssum["sys"].apply(lambda x:x[1:4])
    
    for epocol in ["Epoch"]:
        df_gnssum[epocol] = pd.to_datetime(df_gnssum[epocol])
        
    
    
    ########### TOTSUM
    totsum_stk[1] = _date_format(totsum_stk[1],True)
    
    df_totsum = pd.read_csv(StringIO("\n".join(totsum_stk)),
                            delim_whitespace=True)
    
    for epocol in ["First_Epoch________","Last_Epoch_________"]:
        df_totsum[epocol] = pd.to_datetime(df_totsum[epocol])
        
    df_totsum.drop("#TOTSUM",axis=1,inplace=True)
    
    return df_totsum , df_gnssum


#  _____  _                   
# |  __ \(_)                  
# | |__) |_ _ __   __ _  ___  
# |  _  /| | '_ \ / _` |/ _ \ 
# | | \ \| | | | | (_| | (_) |
# |_|  \_\_|_| |_|\__, |\___/ 
#                  __/ |      
#                 |___/   

def read_ringo_systems(file_in):
    
    F = open(file_in)
    L = F.readlines()
    
    start_pattern = re.compile(r"# (\w+)$")
    end_pattern = re.compile(r"# -{35,}|\Z") 
    
    # Initialize variables to track the current satellite system and its data
    current_system = None
    current_data = []
    
    # Dictionary to store extracted tables
    tables = {}      
            
    # Iterate through the lines of the text
    for line in L:
        # Check if the line matches the start pattern
        start_match = start_pattern.match(line)
        if start_match:
            # Initialize data for the new satellite system
            current_system = start_match.group(1)
        # Check if the line matches the end pattern
        elif end_pattern.match(line):
            # Store the current data for the last satellite system
            if current_system:
                tables[current_system] = current_data
                current_system = None
                current_data = []
        elif current_system:
            # If neither start nor end pattern matches, append the line to the current data
            current_data.append(line)
        else:
            pass
            
    for sys,data in tables.items():
        df = pd.read_csv(StringIO("\n".join(data)),delim_whitespace=True)
        tables[sys] = df.rename(columns={"%":"prn"})
        
    return tables


def read_ringo_qc(file_in):
    F = open(file_in)
    L = F.readlines()
    
    # Define patterns for identifying the start and end of each table
    end_pattern = re.compile(r"# -{35,}|\Z")  # Updated pattern to also match the end of the input
    
    # Define patterns for the "Summary of each satellite system" and "For each satellite" tables
    summary_pattern = re.compile(r"# Summary of each satellite system")
    for_each_satellite_pattern = re.compile(r"# For each satellite")
    
    current_data = []
    
    # Dictionary to store extracted tables
    tables = {}
    
    # Flag to determine if the current section is within "Summary of each satellite system" or "For each satellite"
    in_summary_section = False
    in_for_each_satellite_section = False
    
    # Iterate through the lines of the text
    for iline,line in enumerate(L):
        # Check if the line matches the start pattern
        
        # Check if the line matches the "Summary of each satellite system" pattern
        if summary_pattern.match(line):
            in_summary_section = True
            current_data = []
    
        # Check if the line matches the "For each satellite" pattern
        elif for_each_satellite_pattern.match(line):
            in_for_each_satellite_section = True
            current_data = []
    
        # Check if the line matches the end pattern
        elif end_pattern.match(line) or iline+1 == len(L):
            # If the current section is within "Summary of each satellite system," save the data in the tables dictionary
            if in_summary_section:
                tables["qc_sys"] = current_data
                in_summary_section = False
            # If the current section is within "For each satellite," save the data in the tables dictionary
            elif in_for_each_satellite_section:
                tables["qc_prn"] = current_data
                in_for_each_satellite_section = False
    
        # If the current line contains data, append it to the current_data list
        else:
            current_data.append(line.strip())
            
        
    ### format the table
    for qc,data in tables.items():
        # interpret it as a DataFrame
        df = pd.read_csv(StringIO("\n".join(data)),sep=",")
        df.replace('----',np.nan,regex=True,inplace=True)
        
        # rename the columns
        # sor stands for slips/observation ratio
        cols = [ qc[3:], 'std_mp1', 'std_mp2', 'std_mp5', 'sor_mp1',
                'sor_mp2', 'sor_mp5', 'sor_gf', 'sor_mw', ' sor_iod']
        df.columns = cols
        
        for col in cols[4:]:
            # remove the ratio with / and convert it for a list [slips , nobs]
            df[col] = df[col].str.split("/")
            df[col] = df[col].apply(lambda x: [int(x[0]),int(x[1])] if int(x[1]) != 0 else [np.nan]*2)
            so_stk = np.stack(df[col],dtype=float)
            # extract slips  an nobs in dedicated df columns
            df["slips_"+col.split("_")[1]] = so_stk[:,0]
            df["nobs_"+col.split("_")[1]] = so_stk[:,1]
            # finaly repalce the sor columns with the actual ratio
            df[col] = df[col].apply(lambda x: np.divide(*x))
            
        tables[qc] = df
        
    return tables


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.utils contains functions for misc. 
low level function 

it can be imported directly with:
from geodezyx import utils

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import inspect
import io
#### Import the logger
import logging
import os
import pathlib
import pickle
import re
import sys
import tempfile
import time

import numpy as np
import pathvalidate
import scipy

#### geodeZYX modules
log = logging.getLogger('geodezyx')


##########  END IMPORT  ##########


def clear_all():
    '''Clears all the variables from the workspace of the spyder
    application.'''
    # cls()
    gl = globals().copy()
    for var in gl:
        if var[0] == '_': continue
        if 'func' in str(globals()[var]): continue
        if 'module' in str(globals()[var]): continue

        del globals()[var]


def spyder_run_check():
    """
    Check if the code is run inside Spyder IDE
    """
    if any('SPYDER' in name for name in os.environ):
        return True
    else:
        return False


def is_iterable(inp,
                consider_str_as_iterable=False,
                consider_dict_as_iterable=False):
    """
    Test if the input is an iterable like a list or a numpy array or not

    Parameters
    ----------
    inp : list, numpy.array, ...

    consider_str_as_iterable : bool
        string are considered as iterable by Python per default
        This boolean will avoid True as return if you test a string
        
    Returns
    -------
    out : bool
        True if inp is iterable, False either
    """
    if not consider_str_as_iterable and type(inp) is str:
        return False

    if not consider_dict_as_iterable and type(inp) is dict:
        return False

    try:
        iter(inp)
    except TypeError:
        out = False
    else:
        out = True
    return out


def listify(inp):
    """
    Convert the input into a list.

    Parameters
    ----------
    inp : any
        The input to be converted into a list.

    Returns
    -------
    list
        A list containing the input elements if the input is iterable,
        otherwise a list with the input as its single element.
    """
    if is_iterable(inp):
        return list(inp)
    else:
        return [inp]


def is_not_iterable(inp, consider_str_as_iterable=False):
    """
    Simple negation of is_iterable()
    """
    if not consider_str_as_iterable and type(inp) is str:
        return False

    try:
        iter(inp)
    except TypeError:
        out = True
    else:
        out = False
    return out


def transpose_vector_array(X):
    """
    transpose a Nx3 array to a 3xN array if necessary
    (necessary for some usages)
    
    Parameters
    ----------
    X : iterable
    
    Returns
    -------
    X : iterable
        X transposed if necessary.

    """
    import pandas as pd
    if isinstance(X, pd.core.frame.DataFrame):
        X = X.values

    if isinstance(X, np.ndarray):
        if not X.shape[0] == 3:
            X = X.T
    return X


def is_lambda(v):
    """
    Check if v is lambda

    Source 
    -------
    https://stackoverflow.com/questions/3655842/how-can-i-test-whether-a-variable-holds-a-lambda
    """

    LAMBDA = lambda: 0
    return isinstance(v, type(LAMBDA)) and v.__name__ == LAMBDA.__name__


# def contains_word(s, w):
#     return f' {w} ' in f' {s} '


def docstring_generic():
    """
    prints and returns an prototype generic docstring. Based on Numpy docstring
    convention
    
    Source
    ------
    https://numpydoc.readthedocs.io/en/latest/format.html
    
    """

    docstr_out = """    
    General description

    Parameters
    ----------
    param1 : float or int or str or dict or n-tuple or bool or list or numpy.array
        Description param1

    param2 : float or int or str or dict or n-tuple or bool or list or numpy.array
        Description param2
        
    param3 : float or int or str or dict or n-tuple or bool or list or numpy.array
        Description param3
                
    Returns
    -------
    out1 : float or int or str or dict or n-tuple or bool or list or numpy.array
        Description out1
    
    out2 : float or int or str or dict or n-tuple or bool or list or numpy.array
        Description out2
        
    Note
    ----
    Misc. Notes

    Source
    ------
    www.forum-source.com
    
    Examples
    --------
    >>> answer
    42    
    """

    print(docstr_out)

    return docstr_out


def timeout(func, args=(), kwargs={}, timeout_duration=1, default=None):
    '''
    This function will spwan a thread and run the given function
    using the args, kwargs and return the given default
    value if the timeout_duration is exceeded

    http://stackoverflow.com/questions/366682/how-to-limit-execution-time-of-a-function-call-in-python
    '''
    import threading
    class InterruptableThread(threading.Thread):
        def __init__(self):
            threading.Thread.__init__(self)
            self.result = default

        def run(self):
            try:
                self.result = func(*args, **kwargs)
            except:
                self.result = default

    start = time.time()
    it = InterruptableThread()
    it.start()
    it.join(timeout_duration)
    if it.isAlive():
        log.error('a timout is triggered !!!')
        stop = time.time() - start
        log.error("runtime : %s sec.", stop)
        return it.result
    else:
        stop = time.time() - start
        log.info("runtime : %s sec.", stop)
        return it.result


def indice_printer(i, print_every=10, text_before=''):
    """
    print an index every N iteration
    """
    if np.mod(i, print_every) == 0:
        log.info("%s %s", text_before, i)
    return None


def pickle_saver(datain, outdir=None, outname=None, ext='.pik',
                 timestamp=False, full_path=None):
    """
    Save a Python object in a Pickle file.
    
    Wrapper of pickle.dump
    
    Parameters
    ----------
    datain : generic
        Data which will be saved as a pickle.
    outdir : str, optional
        output directory. The default is None.
    outname : str, optional
        pickle output name. The default is None.
    ext : str, optional
        pickle file extension. The default is '.pik'.
    timestamp : bool, optional
        add the timestamp in the pickle's filename.
        The default is False.
    full_path : str, optional
        gives the full path where to save the pickle.
        if full_path is given, override outdir and outname.
        The default is None.
    Returns
    -------
    outpath : str
        the output pickle file path.

    """

    if not full_path and not outdir and not outname:
        log.error('not full_path and not outdir and not outname are given !')
        raise Exception

    if timestamp:
        ts = get_timestamp() + '_'
    else:
        ts = ''
    if full_path:
        outpath = full_path
    else:
        outpath = os.path.join(outdir, ts + outname + ext)

    # For Python 2
    # pickle.dump( datain , open( outpath , "w+" ) )
    # For Python 3
    pickle.dump(datain, open(outpath, "wb"))
    return outpath


def pickle_loader(pathin):
    """
    Load a Python object saved as a Pickle file.
    
    Wrapper of pickle.load

    Parameters
    ----------
    pathin : str
        the input pickle file path.

    Returns
    -------
    outdata : generic
        Data which will be loaded from the a pickle..

    """

    try:
        # outdata = pickle.load( open( pathin , "r" ) )
        outdata = pickle.load(open(pathin, 'rb'), encoding='latin1')
    except UnicodeDecodeError:
        log.info('alternative reading following a UnicodeDecodeError')
        # outdata = pickle.load( open( pathin ,'rb' )  , encoding='latin1')
        outdata = pickle.load(open(pathin, "r"))
    except TypeError:
        log.info('alternative reading following a TypeError (is it a Py2 pickle ?)')
        # source :
        # https://stackoverflow.com/questions/28218466/unpickling-a-python-2-object-with-python-3
        outdata = pickle.load(open(pathin, 'rb'), encoding='latin1')

    return outdata


def globals_filtered():
    """
    Filter globals() variables with only compatible variables for pickle.
    
    https://stackoverflow.com/questions/2960864/how-to-save-all-the-variables-in-the-current-python-session


    Returns
    -------
    data_out : dict
        filtered globals() variables.

    """
    from spyder_kernels.utils.nsview import globalsfilter, get_supported_types

    data = globals()

    # settings = VariableExplorer.get_settings()

    data_out = globalsfilter(globals(),
                             check_all=True,
                             filters=tuple(get_supported_types()['picklable']),
                             exclude_private=True,
                             exclude_uppercase=False,
                             exclude_capitalized=True,
                             exclude_unsupported=True,
                             excluded_names=[],
                             exclude_callables_and_modules=True)

    return data_out


def detect_encoding(file_path):
    """
    Detect the encoding of a text file.

    This function uses the chardet library to detect the encoding of a given text file. It reads the file line by line and feeds each line to a chardet UniversalDetector. When the detector has made a determination, it stops reading the file and returns the detected encoding.

    Parameters
    ----------
    file_path : str
        The path to the text file for which to detect the encoding.

    Returns
    -------
    str
        The detected encoding of the text file.

    Source
    ------
    https://www.geeksforgeeks.org/detect-encoding-of-a-text-file-with-python/
    """

    import chardet

    with open(file_path, 'rb') as file:
        detector = chardet.universaldetector.UniversalDetector()
        for line in file:
            detector.feed(line)
            if detector.done:
                break
        detector.close()
    return detector.result['encoding']


def open_readlines_smart(file_in, decode_type="iso-8859-1", verbose=False):
    """
    This function takes an input object, opens it, and reads its lines.
    The input file can be the path of a file as a string or as a Path object,
    or the file content as a string, bytes, StringIO object, or a list of lines.

    Parameters
    ----------
    file_in : various
        An input object. This can be a string representing a file path, a Path object,
        a string representing file content, bytes, a StringIO object, or a list of lines.

    decode_type : str, optional
        The decode standard. Default is "iso-8859-1".

    verbose : bool, optional
        If set to True, the function will print the type of the input file. Default is False.

    Returns
    -------
    lines : list
        A list of the lines in the input file.

    Raises
    ------
    FileNotFoundError
        If the file specified by file_in does not exist.
    """

    if verbose:
        log.info("input file is actually a %s", type(file_in))

    if type(file_in) is str and pathvalidate.is_valid_filepath(file_in, platform='auto'):
        if verbose:
            log.info("input file is a string path")
        try:
            FILE = open(file_in)
        except FileNotFoundError as e:
            log.error("%s not found", file_in)
            raise (e)
        lines = FILE.readlines()

    elif type(file_in) is pathlib.Path and pathvalidate.is_valid_filepath(file_in, platform='auto'):
        if verbose:
            log.info("input file is a Path Object")
        try:
            FILE = open(file_in)
        except FileNotFoundError as e:
            log.error("%s not found", file_in)
            raise (e)
        lines = FILE.readlines()

    elif type(file_in) is bytes:
        if verbose:
            log.info("input file is bytes")

        lines = file_in.decode(decode_type).split("\n")

    elif type(file_in) is str:
        if verbose:
            log.info("input file is str")
        lines = file_in.split("\n")

    elif type(file_in) is io.StringIO:
        if verbose:
            log.info("input file is StringIO object")
        lines = file_in.readlines

    elif is_iterable(file_in):
        if verbose:
            log.info("input file is an iterable")
        lines = file_in

    else:
        if verbose:
            log.info("input file is unknown...")
        log.error("something wrong while opening %s", file_in)

    return lines


def split_string_after_n_chars_at_space(s, n):
    """
    Splits a string into substrings of a maximum length of n characters,
    but only splits at spaces and inserts a newline (\n) after each substring.

    Parameters
    ----------
    s : str
        The input string to be split.
    n : int
        The maximum length of each substring.

    Returns
    -------
    str
        The modified string with newlines inserted.
    """
    result = []
    lines = s.split("\n")
    for line in lines:
        words = line.split()
        current_line = []

        for word in words:
            if sum(len(w) for w in current_line) + len(current_line) + len(word) > n:
                result.append(' '.join(current_line))
                current_line = [word]
            else:
                current_line.append(word)

        if current_line:
            result.append(' '.join(current_line))

    return '\n'.join(result)


def add_symbol_to_new_lines(s, symbol='·'):
    """
    Adds a specified symbol to the beginning of each new line in a multi-line string.

    Parameters
    ----------
    s : str
        The input multi-line string.
    symbol : str, optional
        The symbol to add to each new line. Default is '·'.

    Returns
    -------
    str
        The modified string with the symbol added to each new line.
    """
    lines = s.split('\n')
    modified_lines = [symbol + line for line in lines]
    return '\n'.join(modified_lines)


def memmap_from_array(arrin):
    nam = str(np.random.randint(99999)) + '.mmp.tmp'
    path = os.path.join(tempfile.mkdtemp(), nam)
    mmp = np.memmap(path, dtype='float64', mode='w+',
                    shape=arrin.shape)
    mmp[:] = arrin[:]
    return mmp


def mmpa(arrin):
    # just a light wrapper
    return memmap_from_array(arrin)


def read_mat_file(pathin, full=False):
    """ low level reader of a MATLAB mat file """
    import scipy.io as sio
    import os
    fullmat = sio.loadmat(pathin)
    data = fullmat[os.path.basename(pathin).split('.')[0]]
    if full:
        return data, fullmat
    else:
        return data


def save_obj_as_file(objin, pathin, prefix, ext='.exp', suffix=''):
    """
    OLD proto-version of pickle saver DISCONTINUED
    """
    if suffix != '':
        suffix = suffix + '_'
    ts = get_timestamp()
    fullpath = os.path.join(pathin, prefix + '_' + suffix + ts + ext)

    # Python 2
    # pickle.dump( objin, open( fullpath , "w+" ) )
    # Python 3
    pickle.dump(objin, open(fullpath, "wb"))

    return fullpath


def save_array_fast(arrin, outname='',
                    outdir='/home/psakicki/aaa_FOURBI',
                    txt=True):
    import uuid
    if outname == '':
        outname = str(uuid.uuid4())[:8]
    if scipy.sparse.issparse(arrin):
        arrin = arrin.toarray()
    outpath = os.path.join(outdir, outname)
    log.info(outpath)
    if txt:
        np.savetxt(outpath, arrin)
    else:
        np.save(outpath, arrin)
    log.info(outpath)
    return outpath


def join_improved(strseparat, *varsin):
    return strseparat.join([str(v) for v in varsin])


def split_improved(strin, sep_left, sep_right):
    return strin.split(sep_left)[1].split(sep_right)[0]


def diagonalize(x, n=10):
    if is_iterable(x):
        x = list(x) * n
        x = np.expand_dims(np.array(x), 1)
        return np.diag(x[:, 0])
    else:
        return np.diag((np.ones((n, n)) * x)[:, 0])


def read_comments(filein, comment='#'):
    outlist = []
    for l in open(filein):
        if l[0] == comment:
            outlist.append(l[1:-1])
    return outlist


def str2float_smart(str_in):
    try:
        out = float(str_in)
    except ValueError:
        out = np.nan
    return out


def str2int_smart(str_in):
    try:
        out = int(str_in)
    except ValueError:
        out = np.nan
    return out


def str2int_float_autodetect(str_list_in):
    intflt_list_out = []
    for str_in in str_list_in:
        flt_tmp = float(str_in)
        if np.isclose(flt_tmp - np.floor(flt_tmp), 0):
            val = int(flt_tmp)
        else:
            val = flt_tmp
        intflt_list_out.append(val)
    return (intflt_list_out)


def array_from_lists(*listsin):
    """ fonction pour arreter de galerer avec les conversions
    de lists => matrices """
    out = np.vstack(listsin)
    out = out.T
    return out


def stringizer(tupin, separ=' ', eol=True):
    """
    transform elts of a tuple in a string line, ready for write in a file
    """
    if eol:
        end = '\n'
    else:
        end = ''
    lineout = ' '.join([str(e) for e in tupin]) + end
    return lineout


def eval_a_dict(dictin, where, verbose=True):
    """
    where is most of time globals()
    WARN : doesnt work in a function !!!
    use instead :
    for k,v in booldic.items():
        globals()[k] = v
        locals()[k] = v
    """
    for k, v in dictin.items():
        where[k] = v
        if verbose:
            log.info('the 2 values must be equals %s, %s variable', where[k], v, k)
    return None


def boolean_dict(list_of_keywords):
    outdict = dict()
    for k in list_of_keywords:
        outdict[k] = (True, False)
    return outdict


def line_in_file_checker(file_path, string):
    # Open the file
    fobj = open(file_path)

    trigger = False
    for line in fobj:
        if string in line:
            trigger = True
            break

    fobj.close()
    return trigger


def line_count(filein):
    num_lines = sum(1 for line in open(filein))
    return num_lines


def is_in_str(string, *patterns):
    """
    recipe to the famous problem of pattern in string
    from
    http://stackoverflow.com/questions/3389574/check-if-multiple-strings-exist-in-another-string
    """
    L = [x in string for x in patterns]
    return bool(any(L))


def extract_text_between_elements(file_path, elt_start, elt_end):
    """
    source :
        https://stackoverflow.com/questions/9222106/how-to-extract-information-between-two-unique-words-in-a-large-text-file
    """

    f = open(file_path, 'r')
    text = f.read()
    text_extract = text.split(elt_start, 1)[-1].split(elt_end)[0]

    return text_extract


def extract_text_between_elements_2(file_path, elt_start, elt_end,
                                    return_string=False,
                                    nth_occur_elt_start=0,
                                    nth_occur_elt_end=0,
                                    invert=False,
                                    verbose=False):
    """
    This function is based on REGEX (elt_start , elt_end are REGEX)
    and can manage several blocks in the same file

    return_string = True  : returns a string of the matched lines
    return_string = False : returns a list of the matched lines
    invert : exclude text between the pattern
    
    NB : in SINEX context, with "+MARKER", use backslash i.e.
         "\\+MARKER"    
         
    NB2 : think about StingIO for a Pandas DataFrame Handeling
    https://docs.python.org/2/library/stringio.html
    """

    import gzip

    ## 3 possibilities
    # file_path is a path, uncompressed
    # file_path in a path, gzip compressed
    # file_path is already a list of lines

    if type(file_path) is str:  ### case 1 : path compressed
        if file_path[-2:] in ("gz", "GZ"):
            F = gzip.open(file_path, "r+")
            F = [e.decode('utf-8') for e in F]
        else:  ### case 2 : path uncompressed
            try:
                F = open(file_path, "r", encoding="ISO-8859-1")
            except:
                F = open(file_path, "r")
    else:  ### case 3 : already a list of lines
        F = file_path

    out_lines_list = []

    trigger = False
    if invert:
        trigger = not trigger

    last_triggered_line = False

    for line in F:

        if re.search(elt_start, line) and nth_occur_elt_start == 0:
            trigger = True
            if invert:
                trigger = not trigger
        elif re.search(elt_start, line) and nth_occur_elt_start > 0:
            nth_occur_elt_start = nth_occur_elt_start - 1

        if re.search(elt_end, line) and nth_occur_elt_end == 0:
            last_triggered_line = True
        elif trigger and re.search(elt_end, line) and nth_occur_elt_end > 0:
            nth_occur_elt_end = nth_occur_elt_end - 1

        if trigger:
            if verbose:
                log.info(line)
            out_lines_list.append(line)

        if last_triggered_line:
            last_triggered_line = False
            trigger = False
            if invert:
                trigger = not trigger

    if return_string:
        return "".join(out_lines_list)
    else:
        return out_lines_list


def str_2_float_line(line, sep=" ", out_type=float):
    """
    convert a line of number (in str)
    to a list of float (or other out_type)
    """
    fields = line.strip().split(sep)
    fields = [e for e in fields if len(e) > 0]
    try:
        return [out_type(e) for e in fields]
    except:
        return fields


def replace_in_file(file_in, str_before, str_after):
    """
    https://stackoverflow.com/questions/17140886/how-to-search-and-replace-text-in-a-file
    """
    with open(file_in, 'r') as file:
        filedata = file.read()

    # Replace the target string
    filedata = filedata.replace(str_before, str_after)

    # Write the file out again
    with open(file_in + "", 'w') as file:
        file.write(filedata)

    ################
    ### "GETTERs" : wrappers to get easily some fcts
    ################


def get_timestamp(outstring=True, separator='T', utc=False):
    """
    frontend to get easily a timestamp
    """

    if utc:
        now = dt.datetime.now(dt.timezone.utc)
        zulu = "Z"
    else:
        now = dt.datetime.now()
        zulu = ""

    if outstring:
        return now.strftime('%Y%m%d' + separator + '%H%M%S') + zulu
    else:
        return now


def get_function_name():
    return inspect.stack()[0][3]


def get_computer_name():
    import platform
    return platform.node()


def get_username():
    import pwd
    return pwd.getpwuid(os.getuid())[0]


def vectorialize(array_in):
    """
    redondant avec .flat ???
    """
    vector_out = array_in.reshape((array_in.size,))
    return vector_out


def get_specific_locals(prefix):
    """ get locals params with 'prefix' in the name
        can actually be a suffix """
    # MARCHE PAS EN L'ETAT
    loctemp = dict(globals())
    log.info(loctemp)
    outlis = []
    for k, v in loctemp.items():
        if prefix in k:
            outlis.append(str(k[4:]) + ' : ' + str(v))
    outlis.sort()
    return outlis


def get_type_smart(obj_in):
    """
    get type of an object, to convert easily another one to this type
    for instance type(np.array(A)) doesn't return a constructor 
    """
    typ = type(obj_in)
    if typ is np.ndarray:
        return np.array
    else:
        return typ


class Tee(object):
    """
    Internal class for Tee_frontend
    
    Source
    ------
        based on
        http://stackoverflow.com/questions/11325019/output-on-the-console-and-file-using-python
        Secondary links
        http://stackoverflow.com/questions/616645/how-do-i-duplicate-sys-stdout-to-a-log-file-in-python
        http://stackoverflow.com/questions/2996887/how-to-replicate-tee-behavior-in-python-when-using-subprocess
    """

    def __init__(self, *files):
        self.original = sys.stdout
        self.files = [sys.stdout] + list(files)
        sys.stdout = self

    def write(self, obj):
        for f in self.files:
            f.write(obj)

    def flush(self):
        for f in self.files:
            f.flush()

    def stop(self):
        for fil in self.files:
            if fil != self.original:
                fil.close()
        sys.stdout = self.original
        self.files = [self.original]

    def pause(self):
        sys.stdout = self.original
        self.files_saved = list(self.files)
        self.files = [self.original]

    def restart(self):
        self.files = self.files_saved
        sys.stdout = self


def Tee_frontend(dir_in, logname_in, suffix='', ext='log', print_timestamp=True):
    """
    Write in a file the console output

    Parameters
    ----------
    dir_in : str
        directory path.
    logname_in : str
        logfile name.
    suffix : str, optional
        An optional suffix. The default is ''.
    ext : str, optional
        file extension. The default is 'log'.
    print_timestamp : bool, optional
        print a timestamp in the filename. The default is True.

    Returns
    -------
    F_tee : F_tee object
        Object controling the output
        
    Note
    ----
        It is recommended to stop the writing at the end of the script
        with F_tee.stop()

    """
    # if suffix != '':
    #    suffix = suffix + '_'
    if print_timestamp:
        ts = '_' + get_timestamp()
    else:
        ts = ''
    f = open(os.path.join(dir_in, logname_in + '_' + suffix + ts + '.' + ext), 'w')
    F_tee = Tee(f)
    # sys.stdout = f_tee
    return F_tee


def alphabet(num=None):
    if num is None:
        import string
        return list(string.ascii_lowercase)
    else:
        return alphabet()[num]


def alphabet_reverse(letter=None):
    if letter is None:
        import string
        return list(string.ascii_lowercase)
    else:
        return alphabet_reverse().index(letter)


def dday():
    """
    Give the time span between present and toolbox author's PhD defense date
    
    (tests also the console messages)
        
    Note
    ----
    https://fr.wiktionary.org/wiki/quille#Dérivés_2
    
    https://en.wiktionary.org/wiki/quille

    Returns
    -------
    D : datetime
        elapsed time.

    """
    D = (dt.datetime(2016, 10, 14) - dt.datetime.now()).days
    print("used logger:", log)

    from logging_tree import printout
    printout()

    log.warning('J - %s avant la quille !!!', D)
    log.info('J - %s avant la quille', D)
    log.debug('J - %s avant la quille, chill', D)

    return D


def Aformat(A, landscape=True):
    LA = [(841, 1189),
          (594, 841),
          (420, 594),
          (297, 420),
          (210, 297),
          (148, 210),
          (105, 148),
          (74, 105),
          (52, 74)]

    out = np.array(LA[A]) / 25.4
    if landscape:
        out = np.flipud(out)
    return tuple(out)


def greek_alphabet(num=None, maj=False):
    if not num:
        a = ['\u03B1',
             '\u03B2',
             '\u03B3',
             '\u03B4',
             '\u03B5',
             '\u03B6',
             '\u03B7',
             '\u03B8',
             '\u03B9',
             '\u03BA',
             '\u03BB',
             '\u03BC',
             '\u03BD',
             '\u03BE',
             '\u03BF',
             '\u03C0',
             '\u03C1',
             '\u03C3',
             '\u03C4',
             '\u03C5',
             '\u03C6',
             '\u03C7',
             '\u03C8',
             '\u03C9']

        A = ['\u0391',
             '\u0392',
             '\u0393',
             '\u0394',
             '\u0395',
             '\u0396',
             '\u0397',
             '\u0398',
             '\u0399',
             '\u039A',
             '\u039B',
             '\u039C',
             '\u039D',
             '\u039E',
             '\u039F',
             '\u03A0',
             '\u03A1',
             '\u03A3',
             '\u03A4',
             '\u03A5',
             '\u03A6',
             '\u03A7',
             '\u03A8',
             '\u03A9']

        if maj:
            return A
        else:
            return a
    else:
        return greek_alphabet()[num]


def trunc(f, n):
    '''Truncates/pads a float f to n decimal places without rounding'''
    slen = len('%.*f' % (n, f))
    return float(str(f)[:slen])


def multidot(tupin):
    out = np.eye(tupin[0].shape[0])

    for e in tupin:
        out = np.dot(out, e)

    return out


def mdot(*args):
    ret = args[0]
    for a in args[1:]:
        ret = np.dot(ret, a)
    return ret


def mdotr(*args):
    ret = args[-1]
    for a in reversed(args[:-1]):
        ret = np.dot(a, ret)
    return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Télécharge (si besoin) les grilles VMF1 puis calcule les mapping functions VMF1
pour DUNQ00FRA au 2021-07-10T01:00:18, en utilisant:
  - download_VMF.download_VMF (ton script de téléchargement)
  - read_vmf1_grid.read_vmf1_grid
  - vmf1_ht.vmf1_ht
Exécuter depuis la racine du projet.
"""

import os
from math import floor
from datetime import datetime, timezone, timedelta
import numpy as np

# --- briques du projet ---
import download_VMF as dl
import read_vmf1_grid as rvg
import vmf1_ht

# --- Patch de portabilité: read_vmf1_grid construit des chemins avec '\\'.
#     On monkey-patch "open" dans le module pour convertir en séparateur OS.
import builtins
rvg.open = lambda path, *args, **kwargs: builtins.open(path.replace("\\", os.sep), *args, **kwargs)

# ---------------- outils temps ----------------
def iso_to_mjd(iso_utc: str) -> float:
    """Convertit une date ISO UTC en MJD (UTC)."""
    dt = datetime.fromisoformat(iso_utc)
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    else:
        dt = dt.astimezone(timezone.utc)
    Y, M, D = dt.year, dt.month, dt.day
    frac_day = (dt.hour + dt.minute/60 + dt.second/3600 + dt.microsecond/3.6e9)/24.0
    if M <= 2:
        Yp, Mp = Y - 1, M + 12
    else:
        Yp, Mp = Y, M
    A = floor(Yp/100)
    B = 2 - A + floor(A/4)
    JD = floor(365.25*(Yp+4716)) + floor(30.6001*(Mp+1)) + (D + frac_day) + B - 1524.5
    return JD - 2400000.5

def six_hour_bracketing_dates(dt_utc: datetime):
    """Retourne l'ensemble des dates (YYYY,MM,DD) impliquées par l'interpolation 6h autour de dt."""
    base = dt_utc.replace(minute=0, second=0, microsecond=0)
    # borne basse (multiple de 6h)
    floor6 = base.replace(hour=(base.hour//6)*6)
    ceil6  = floor6 + timedelta(hours=6)
    return {
        (floor6.year, floor6.month, floor6.day),
        (ceil6.year,  ceil6.month,  ceil6.day)
    }

# ---------------- téléchargement si besoin ----------------
def ensure_vmf1_files_for_epoch(dt_utc: datetime, vmf_dir: str):
    """
    Vérifie la présence des grilles nécessaires dans vmf_dir.
    Si manquantes, appelle download_VMF.download_VMF(year, month, day) pour chaque date utile.
    """
    os.makedirs(vmf_dir, exist_ok=True)

    needed_dates = six_hour_bracketing_dates(dt_utc)  # au pire 2 dates (avant/après)
    # On considère que la fonction de lecture peut avoir besoin des H00/H06/H12/H18 de ces dates.
    # On télécharge donc la journée complète pour chaque date impliquée (c'est ainsi que marche ton downloader).
    for (y, m, d) in sorted(needed_dates):
        # fichiers attendus
        wanted = [
            f"VMFG_{y}{str(m).zfill(2)}{str(d).zfill(2)}.H{hh}"
            for hh in ("00", "06", "12", "18")
        ]
        have = all(os.path.exists(os.path.join(vmf_dir, w)) for w in wanted)
        if not have:
            print(f"[info] Grilles VMF1 manquantes pour {y}-{str(m).zfill(2)}-{str(d).zfill(2)}. "
                  f"Téléchargement…")
            # Le downloader télécharge dans le répertoire courant: on se place donc dans vmf_dir le temps du DL.
            cwd = os.getcwd()
            try:
                os.chdir(vmf_dir)
                # Très important: ton download_VMF attend un month *tel quel* dans le nom du fichier.
                # On lui passe une chaîne zero-padded pour garantir 'YYYYMMDD'.
                dl.download_VMF(year=str(y), month=str(m).zfill(2), day=d)
            finally:
                os.chdir(cwd)
        else:
            print(f"[ok] Grilles VMF1 déjà présentes pour {y}-{str(m).zfill(2)}-{str(d).zfill(2)}.")

# ---------------- paramètres station & epoch ----------------
station = {
    "name": "DUNQ00FRA",
    "lon_deg": 2.366699,
    "lat_deg": 51.048100,
    "h_m": 51.111900,
}
epoch_iso_utc = "2021-07-10T01:00:18"  # UTC
elevations_deg = [5, 10, 15, 20, 30, 45, 60]  # à adapter si besoin

def main():
    # chemins
    project_root = os.getcwd()
    vmf_dir = os.path.join(project_root, "Mapping_Fcn", "vmf1")

    # 1) S'assurer que les grilles nécessaires sont présentes (sinon les télécharger)
    dt = datetime.fromisoformat(epoch_iso_utc).replace(tzinfo=timezone.utc)
    ensure_vmf1_files_for_epoch(dt, vmf_dir)

    # 2) Temps → MJD
    mjd = iso_to_mjd(epoch_iso_utc)

    # 3) Lire coefficients a_h, a_w à la station/datetime
    ah_vec, aw_vec = rvg.read_vmf1_grid(
        mjd=[mjd],
        ell=[station["lat_deg"], station["lon_deg"], station["h_m"]]
    )
    ah = float(np.atleast_1d(ah_vec).ravel()[0])
    aw = float(np.atleast_1d(aw_vec).ravel()[0])

    # 4) Calculer VMF1 (hydro & humide) pour une liste d’élévations
    dlat_rad = np.deg2rad(station["lat_deg"])
    ht_m = station["h_m"]

    print(f"# Station   : {station['name']}")
    print(f"# Lat, Lon  : {station['lat_deg']:.6f}°, {station['lon_deg']:.6f}°")
    print(f"# Height    : {ht_m:.3f} m")
    print(f"# Epoch UTC : {epoch_iso_utc}  (MJD {mjd:.6f})\n")
    print(f"ah (grid) = {ah:.6e}")
    print(f"aw (grid) = {aw:.6e}\n")

    print("Elevation(deg),  VMF1_hydro,  VMF1_wet")
    for el_deg in elevations_deg:
        zd_rad = np.deg2rad(90.0 - el_deg)  # distance zénithale
        vmf1h, vmf1w = vmf1_ht.vmf1_ht(
            ah=ah, aw=aw, dmjd=mjd, dlat=dlat_rad, ht=ht_m, zd=zd_rad
        )
        print(f"{el_deg:>6.1f}         {vmf1h:>10.6f}  {vmf1w:>10.6f}")

if __name__ == "__main__":
    main()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 26/02/2025 11:36:09

@author: psakic
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
#### Import the logger
import logging
import os
import subprocess

#### geodeZYX modules
from geodezyx import conv
from geodezyx import files_rw
from geodezyx import operational
from geodezyx import utils

import geodezyx.operational.gins_runner.gins_common as ginscmn

log = logging.getLogger('geodezyx')

def sp3_2_gins(sp3_pathin):
    os.chdir(os.path.dirname(sp3_pathin))
    # print("WARN : the path of the GINS conversion tool has been hardcoded !!!")
    kommand_1 = ginscmn.get_gin_path() + "/gins_toolbox/scripts/"
    log.info(kommand_1)
    kommand = kommand_1 + "sp3-gins " + sp3_pathin
    #    stream = os.popen(kommand) # forrtl: severe if we use os.popen
    #    subprocess.call([kommand], shell=True)
    log.info(kommand)
    p = subprocess.Popen(
        [kommand], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out, err = p.communicate()
    return sp3_pathin + ".gin"


def clk_2_gins(clk_pathin):
    os.chdir(os.path.dirname(clk_pathin))
    # print("WARN : the path of the GINS conversion tool has been hardcoded !!!")
    kommand_1 = ginscmn.get_gin_path() + "/gins_toolbox/scripts/"
    log.info(kommand_1)
    kommand = kommand_1 + "clk2gins.sh " + os.path.basename(clk_pathin)
    # kommand = kommand_1 + "clk2gins.sh " + clk_pathin
    #    stream = os.popen(kommand) # forrtl: severe if we use os.popen
    #    subprocess.call([kommand], shell=True)
    log.info(kommand)
    p = subprocess.Popen(
        [kommand], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out, err = p.communicate()
    return clk_pathin + ".gins"


def sort_orbit_gins(pathin, pathout):
    kommand = "sort -n -k1,2 -k3 " + pathin + " > " + pathout
    #    stream = os.popen(kommand) # forrtl: severe if we use os.popen
    #    subprocess.call([kommand], shell=True)
    p = subprocess.Popen(
        [kommand], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out, err = p.communicate()
    return pathout


def bad_sat_finder(orbfilein, egrep_ready=True):
    satdic = dict()
    for l in open(orbfilein):
        f = l.split()
        satid = int(f[0])
        if satid not in satdic:
            satdic[satid] = []
        D = conv.jjul_cnes2dt(int(f[1])) + dt.timedelta(seconds=float(f[2]))
        satdic[satid].append(D)

    epochdic = dict()
    epochlis = []
    for k, v in satdic.items():
        epochdic[k] = len(v)
        epochlis.append(len(v))

    nominal_epoch = utils.most_common(epochlis)

    bad_sat_lis = []

    for k, v in epochdic.items():
        if v != nominal_epoch:
            bad_sat_lis.append(k)

    if egrep_ready:
        bad_sat_lis = ["^  " + str(e) for e in bad_sat_lis]

    return bad_sat_lis


def orbit_cleaner(orbfilein, orbfileout):
    """remove sat of a gins orbit file without a nominal number of epoch"""
    bad_sat_lis = bad_sat_finder(orbfilein)
    goodsatslinlis = utils.grep(orbfilein, bad_sat_lis, invert=1, regex=1)

    filobj = open(orbfileout, "w+")

    for lin in goodsatslinlis:
        filobj.write(lin)

    filobj.close()

    return bad_sat_lis


def download_convert_2_gins_orb_clk(
    centraldate,
    work_folder=None,
    ac="jpl",
    repro=2,
    rm_temp_files=True,
    data_center="cddis",
    force=False,
):

    temp_dir = work_folder

    if not temp_dir:
        temp_dir = os.path.join(ginscmn.get_gin_path(), "gin", "TEMP_DATA")
        log.info("use default forder %s", temp_dir)
    if not os.path.exists(temp_dir):
        os.makedirs(temp_dir)

    strtdt = centraldate - dt.timedelta(days=1)
    enddt = centraldate + dt.timedelta(days=1)

    if centraldate > dt.datetime(2013, 12, 29) and repro != 0:
        log.warning(str(centraldate) + " a bit late for repro " + str(repro))

    # defining names & paths
    if repro != 0:
        ac_prefix = ac[0:2] + str(repro)  # os.path.basename(gins_orb_lis[0])[0:3]

    else:
        ac_prefix = ac

    week, dow = conv.dt2gpstime(centraldate)
    week, dow = str(week), str(dow)
    centdate = week + dow + centraldate.strftime("_%Y_%j")
    catoutorb = ac_prefix + centdate + ".3d.orb.gin"
    sortoutorb = ac_prefix + centdate + ".3d.orb.sort.gin"
    cleanoutorb = ac_prefix + centdate + ".3d.orb.clean.gin"

    catoutclk = ac_prefix + centdate + ".3d.clk.gin"

    catoutorb_path = os.path.join(temp_dir, catoutorb)
    sortoutorb_path = os.path.join(temp_dir, sortoutorb)
    cleanoutorb_path = os.path.join(temp_dir, cleanoutorb)
    catoutclk_path = os.path.join(temp_dir, catoutclk)

    # preliminary check, if the orbits already exist
    if (
        os.path.isfile(cleanoutorb_path)
        and os.path.isfile(catoutclk_path)
        and not force
    ):
        print("INFO : converted orbits & clocks already exist, it's nice")
        return cleanoutorb_path, catoutclk_path

    # downloading orbits
    sp3_zlis = operational.multi_downloader_orbs_clks_2(
        temp_dir,
        strtdt,
        enddt,
        prod_types=("sp3",),
        AC_names=(ac,),
        repro=repro,
        archtype="/",
        sorted_mode=0,
        data_center=data_center,
    )
    # sorted_mode must be off !!!
    # elsewhere the cat of the orbits/clock is bad !!!

    # downloading clocks
    # 30sec clock prioritary
    clk_zlis = operational.multi_downloader_orbs_clks_2(
        temp_dir,
        strtdt,
        enddt,
        prod_types=("clk",),
        AC_names=(ac,),
        repro=repro,
        archtype="/",
        sorted_mode=0,
        data_center=data_center,
    )
    # sorted_mode must be off !!!
    # elsewhere the cat of the orbits/clock is bad !!!

    # standard clock else
    if not clk_zlis:
        print("INFO : clock download : no 30s clock found, trying std. clocks")
        clk_zlis = operational.multi_downloader_orbs_clks(
            temp_dir,
            strtdt,
            enddt,
            sp3clk="clk",
            ac=ac,
            data_center=data_center,
            repro=repro,
            archtype="/",
            sorted_mode=0,
        )
        # sorted_mode must be off !!!
        # elsewhere the cat of the orbits/clock is bad !!!

    # Exception case where no files are found
    if len(sp3_zlis) < 3 or len(clk_zlis) < 3:
        print(
            "ERR : download_convert_2_gins_orb_clk : missing sp3/clk for", centraldate
        )
        print("list of sp3 : ", sp3_zlis)
        print("list of clk : ", clk_zlis)
        return None, None

    # uncompressing orbits
    sp3lis = []
    for Zfil in sp3_zlis:
        if Zfil.endswith(".Z") or Zfil.endswith(".gz"):
            uncomp_fil = files_rw.unzip_gz_z(Zfil)
        else:
            uncomp_fil = Zfil
        # some SP3 haven't EOF at the end ...
        if not utils.grep_boolean(uncomp_fil, "EOF"):
            with open(uncomp_fil, "a") as myfile:
                print("INFO : uncompress SP3 : adding a EOF line")
                myfile.write("EOF")
                myfile.close()

        sp3lis.append(uncomp_fil)

    # uncompressing clocks
    clklis = []
    for Zfil in clk_zlis:
        if Zfil.endswith(".Z") or Zfil.endswith(".gz"):
            uncomp_fil = files_rw.unzip_gz_z(Zfil)
        else:
            uncomp_fil = Zfil
        clklis.append(uncomp_fil)

    # converting orbits
    gins_orb_lis = []
    for sp3 in sp3lis:
        gins_orb_lis.append(sp3_2_gins(sp3))

    # converting clocks
    gins_clk_lis = []
    for clk in clklis:
        gins_clk_lis.append(clk_2_gins(clk))

    # cating orbits
    utils.cat(catoutorb_path, *gins_orb_lis)
    sort_orbit_gins(catoutorb_path, sortoutorb_path)

    # test, checking if the cat was OK
    catoutclk_nline = utils.line_count(sortoutorb_path)
    if catoutclk_nline < 7500:
        print("WARN : the 3days orb file (~8000lines) looks small ! ")
        print(catoutclk_nline, "lines found in", sortoutorb_path)

    # cleaning the orbits file
    _ = orbit_cleaner(sortoutorb_path, cleanoutorb_path)
    Nsort = utils.line_count(sortoutorb_path)
    Nclean = utils.line_count(cleanoutorb_path)

    badsatlis = bad_sat_finder(sortoutorb_path, 0)

    if len(badsatlis) != 0:
        print("INFO : ", len(badsatlis), "bad sats.", Nsort - Nclean, "epochs del.")
        print("       bad sats :", badsatlis)

    # cating clocks
    utils.cat(catoutclk_path, *gins_clk_lis)

    if rm_temp_files:
        try:
            os.remove(catoutorb_path)
            os.remove(sortoutorb_path)
            [os.remove(e) for e in sp3lis]
            [os.remove(e) for e in clklis]
            [os.remove(e) for e in gins_orb_lis]
            [os.remove(e) for e in gins_clk_lis]
        except:
            pass
    return cleanoutorb_path, catoutclk_path


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 25 23:13:15 2023

@author: psakicki
"""

import netCDF4
import numpy as np
import matplotlib.pyplot as plt


####### IFREMER 
p = "./netcdf_exemples/104577.nc"
ncfile_ifremer = netCDF4.Dataset(p,mode='r',format='NETCDF4_CLASSIC')
print(ncfile_ifremer)
depth_kw = "depth" # NOAA
depth_kw = "PRES"
time = np.array(ncfile_ifremer["TIME"][:])
pres = np.array(ncfile_ifremer["PRES"][:])
pres = pres.squeeze()
time = time.squeeze()
plt.plot(time,pres)
ncfile_ifremer["LATITUDE"][:]

####### NOAA
p = "./netcdf_exemples/wc32_19910624to19920604.nc"
ncfile_noaa = netCDF4.Dataset(p,mode='r',format='NETCDF4_CLASSIC')
print(ncfile_noaa)

ncfile_noaa["timeSeries"]
ncfile_noaa["pressure_sensor_info"]

####### HOMEMADE
p_homemade = "/home/psakicki/CODES/geodezyx/geodezyx/marine/netcdf_a0a/output_netcdf/out.nc"
ncfile_homemade = netCDF4.Dataset(p_homemade,mode='r',format='NETCDF4_CLASSIC')
print(ncfile_homemade)



from .anubis_frontend import *
from .cluster_gfz import *

# from .gnss_downloader     import *
from .gins_runner import *
from .bin_opera import *
from .download_cddis import *
from .download_dropbox import *
from .download_find_files import *
from .download_prods import *
from .download_rinex import *
from .download_rinex_legacy import *
from .download_utils import *
from .groops_frontend import *
from .hector_frontend import *
from .midas_frontend import *
from .pride_pppar_frontend import *
from .rinex_lister_plotter import *
from .rinex_utils import *
from .rtklib_frontend import *
from .track_frontend import *
